
###############################################################################
# Base object
###############################################################################


class item():
    '''
    unique_id = parser : item
    '''

    def __init__(self, sString):
        self.value = sString
        self.indent = None
        self.hierarchy = None
        self.context = []
        self.code_tags = []
        self.base_token, self.sub_token = self.update_token_types()
        self.filename = None

    def update_token_types(self):
        try:
            lDoc = self.__doc__.split()
            for iDoc, sDoc in enumerate(lDoc):
                if sDoc == 'unique_id':
                    return lDoc[iDoc + 2], lDoc[iDoc + 4]
            return None, None
        except AttributeError:
            return None, None

    def get_value(self):
        return self.value

    def set_value(self, sString):
        self.value = sString

    def length(self):
        return len(self.value)

    def set_indent(self, iIndent):
        self.indent = iIndent

    def get_indent(self):
        return self.indent

    def set_hierarchy(self, iHierarchy):
        self.hierarchy = iHierarchy

    def get_hierarchy(self):
        return self.hierarchy

    def add_context(self, sContext):
        self.context.extend(sContext)

    def pop_context(self):
        return self.context.pop()

    def get_context(self):
        return self.context

    def set_code_tags(self, lCodeTags):
        self.code_tags = lCodeTags.copy()

    def has_code_tag(self, sCodeTag):
        if self.code_tags == ['all']:
            return True
        if sCodeTag in self.code_tags:
            return True
        return False

    def clear_code_tags(self):
        self.code_tags = []

    def set_all_code_tags(self):
        self.code_tags = ['all']

    def get_unique_id(self, sJoin=None):
        if sJoin is None:
            return self.base_token, self.sub_token
        return self.base_token + sJoin + self.sub_token

    def get_base_token(self):
        return self.base_token

    def get_sub_token(self):
        return self.sub_token

    def set_filename(self, sFilename):
        self.filename = sFilename

    def get_filename(self):
        return self.filename


class todo(item):
    '''
    unique_id = parser : todo
    '''

    def __init__(self, sString):
        item.__init__(self, sString)

###############################################################################
# Base VHDL type classes
###############################################################################

class preprocessor(item):
    '''
    unique_id = parser : preprocessor
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class type(item):
    '''
    unique_id = parser : type
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class function(item):
    '''
    unique_id = parser : function
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class tic(item):
    '''
    unique_id = parser : tic
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class event_keyword(item):
    '''
    unique_id = parser : event_keyword
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class bar(item):
    '''
    unique_id = parser : bar
    '''

    def __init__(self, sString='|'):
        item.__init__(self, '|' )


class open_bracket(item):
    '''
    unique_id = parser : open_bracket
    '''

    def __init__(self, sString='['):
        item.__init__(self, '[')


class close_bracket(item):
    '''
    unique_id = parser : close_bracket
    '''

    def __init__(self, sString=']'):
        item.__init__(self, ']')


class question_mark(item):
    '''
    unique_id = parser : question_mark
    '''

    def __init__(self):
        item.__init__(self, '?')


class undefined_range(item):
    '''
    unique_id = parser : undefined_range
    '''

    def __init__(self):
        item.__init__(self, '<>')


class error(item):
    '''
    unique_id = parser : error
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class carriage_return(item):
    '''
    unique_id = parser : carriage_return
    '''

    def __init__(self):
        item.__init__(self, '\n')


class blank_line(item):
    '''
    unique_id = parser : blank_line
    '''

    def __init__(self):
        item.__init__(self, '')


class none(item):
    '''
    unique_id = parser : none
    '''

    def __init__(self):
        item.__init__(self, None)


class keyword(item):
    '''
    unique_id = parser : keyword
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class identifier(item):
    '''
    unique_id = parser : identifier
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class designator(item):
    '''
    unique_id = parser : designator
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class colon(item):
    '''
    unique_id = parser : colon
    '''

    def __init__(self, sString=':'):
        item.__init__(self, sString)


class comma(item):
    '''
    unique_id = parser : comma
    '''

    def __init__(self, sString=','):
        item.__init__(self, sString)


class dot(item):
    '''
    unique_id = parser : dot
    '''

    def __init__(self, sString='.'):
        item.__init__(self, sString)


class semicolon(item):
    '''
    unique_id = parser : semicolon
    '''

    def __init__(self, sString=';'):
        item.__init__(self, sString)


class whitespace(item):
    '''
    unique_id = parser : whitespace
    '''

    def __init__(self, sString):
        item.__init__(self, sString)
        self.has_tab = False


class comment(item):
    '''
    unique_id = parser : comment
    '''

    def __init__(self, sString):
        item.__init__(self, sString)
        self.is_block_comment = False
        self.block_comment_indent = None
        self.has_tab = False


class logical_name(item):
    '''
    unique_id = parser : logical_name
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class selected_name(item):
    '''
    unique_id = parser : selected_name
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class name(item):
    '''
    unique_id = parser : name
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class simple_name(item):
    '''
    unique_id = parser : simple_name
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class subtype_indication(item):
    '''
    unique_id = parser : subtype_indication
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class condition(item):
    '''
    unique_id = parser : condition
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class expression(item):
    '''
    unique_id = parser : expression
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class static_expression(expression):
    '''
    unique_id = parser : static_expression
    '''

    def __init__(self, sString):
        expression.__init__(self, sString)


class label(item):
    '''
    unique_id = parser : label
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class label_colon(colon):
    '''
    unique_id = parser : label_colon
    '''

    def __init__(self):
        colon.__init__(self)


class open_parenthesis(item):
    '''
    unique_id = parser : open_parenthesis
    '''

    def __init__(self):
        item.__init__(self, '(')


class close_parenthesis(item):
    '''
    unique_id = parser : close_parenthesis
    '''

    def __init__(self):
        item.__init__(self, ')')


class equal_sign(item):
    '''
    unique_id = parser : equal_sign
    '''

    def __init__(self):
        item.__init__(self, '=')


class character_literal(item):
    '''
    unique_id = parser : character_literal
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class string_literal(item):
    '''
    unique_id = parser : string_literal
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class operator_symbol(string_literal):
    '''
    unique_id = parser : operator_symbol
    '''

    def __init__(self, sString):
        string_literal.__init__(self, sString)


class signature(item):
    '''
    unique_id = parser : signature
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class type_mark(item):
    '''
    unique_id = parser : type_mark
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class subtype_definition(item):
    '''
    unique_id = parser : subtype_definition
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class target(item):
    '''
    unique_id = parser : target
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class assignment(item):
    '''
    unique_id = parser : assignment
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class choices(item):
    '''
    unique_id = parser : choices
    '''

    def __init__(self, sString):
        item.__init__(self, sString)


class beginning_of_file(item):
    '''
    unique_id = parser : choices
    '''

    def __init__(self):
        item.__init__(self, 'beginning_of_file')


class slice_name():
    '''
    unique_id = parser : slice_name
    '''

    def __init__(self, lTokens):
        self.tokens = lTokens

    def get_tokens(self):
        return self.tokens


class integer(item):
    '''
    unique_id = parser : integer
    '''

    def __init__(self, sString):
        item.__init__(self, sString)




from vsg.rules import align_tokens_in_region_between_tokens_unless_between_tokens as Rule

from vsg import token

lAlign = []
lAlign.append(token.file_declaration.colon)
lAlign.append(token.variable_declaration.colon)
lAlign.append(token.constant_declaration.colon)
lAlign.append(token.alias_declaration.colon)
lAlign.append(token.alias_declaration.is_keyword)

oStartToken = token.subprogram_body.is_keyword
oEndToken = token.subprogram_body.begin_keyword

lUnless = []
lUnless.append([token.procedure_specification.procedure_keyword, token.subprogram_body.semicolon])


class rule_012(Rule):
    '''
    This rule checks the colons are in the same column for all declarations in the function declarative part.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       variable var1 : natural;
       variable var2  : natural;
       constant c_period : time;

    **Fix**

    .. code-block:: vhdl

       variable var1     : natural;
       variable var2     : natural;
       constant c_period : time;
    '''

    def __init__(self):
        Rule.__init__(self, 'function', '012', lAlign, oStartToken, oEndToken, lUnless)
        self.solution = 'Align :.'
        self.subphase = 2




from vsg.vhdlFile.classify import sequential_statement


def detect(iToken, lObjects):
    '''
    sequence_of_statements ::=
        { sequential_statement }
    '''

    iLast = 0
    iCurrent = iToken
    while iLast != iCurrent:
        iLast = iCurrent
        iCurrent = sequential_statement.detect(iCurrent, lObjects)
    return iCurrent




import shutil

from . import config
from . import junit
from . import rule_list
from . import utils
from . import vhdlFile

from .exceptions import ClassifyError
from .exceptions import ConfigurationError


def create_backup_file(sFileName):
    '''Copies existing file and adds .bak to the end.'''
    shutil.copy2(sFileName, sFileName + '.bak')


def configure_rules(oConfig, oRules, configuration, iIndex, sFileName):

    oRules.configure(oConfig)
    if is_filename_in_file_list(configuration, sFileName):
        iMyIndex = get_index_of_filename_in_file_list(configuration, sFileName)
        oRuleConfig = config.config()
        if does_file_have_rule_configuration(configuration, iMyIndex, sFileName):
            oRuleConfig.dConfig = configuration["file_list"][iMyIndex][sFileName]
            oRules.configure(oRuleConfig)


def is_filename_in_file_list(configuration, sFileName):
    try:
        lFileNames = utils.extract_file_names_from_file_list(configuration["file_list"])
        if sFileName in lFileNames:
            return True
        return False
    except KeyError:
        return False


def get_index_of_filename_in_file_list(configuration, sFileName):
    lFileNames = utils.extract_file_names_from_file_list(configuration["file_list"])
    return lFileNames.index(sFileName)


def does_file_have_rule_configuration(configuration, iMyIndex, sFileName):
    try:
        sTemp = configuration["file_list"][iMyIndex][sFileName]
        return True
    except:
        return False


# This function is in a separate module from __main__ as a workaround for https://bugs.python.org/issue25053
# see also https://stackoverflow.com/questions/41385708/multiprocessing-example-giving-attributeerror/42383397#42383397
def apply_rules(commandLineArguments, oConfig, tIndexFileName):
    configuration = oConfig.dConfig

    dIndent = oConfig.dIndent
    fix_only = oConfig.dFixOnly

    iIndex, sFileName = tIndexFileName
    dJsonEntry = {}
    lFileContent, eError = vhdlFile.utils.read_vhdlfile(sFileName)
    try:
        oVhdlFile = vhdlFile.vhdlFile(lFileContent, sFileName, eError)
    except ClassifyError as e:
        fExitStatus = True
        testCase = create_junit_testcase(sFileName, e)
        dJsonEntry["file_path"] = sFileName
        dJsonEntry["violations"] = []
        sOutputStd = ''
        sOutputErr = e.message
        return fExitStatus, testCase, dJsonEntry, sOutputStd, sOutputErr

    oVhdlFile.set_indent_map(dIndent)
    try:
        oRules = rule_list.rule_list(
            oVhdlFile, oConfig.severity_list, commandLineArguments.local_rules
        )
    except OSError as e:
        sOutputStd = (
            f"ERROR: encountered {e.__class__.__name__}, {e.args[1]} "
            + commandLineArguments.local_rules
            + " when trying to open local rules file."
        )
        sOutputErr = None
        return 1, None, dJsonEntry, sOutputStd, sOutputErr

    try:
        configure_rules(oConfig, oRules, configuration, iIndex, sFileName)
    except ConfigurationError as e:
        fExitStatus = True
        testCase = None
        dJsonEntry["file_path"] = sFileName
        dJsonEntry["violations"] = []
        sOutputStd = ''
        sOutputErr = e.message
        return fExitStatus, testCase, dJsonEntry, sOutputStd, sOutputErr

    if commandLineArguments.fix:
        if commandLineArguments.backup:
            create_backup_file(sFileName)
        oRules.fix(
            commandLineArguments.fix_phase, commandLineArguments.skip_phase, fix_only
        )
        write_vhdl_file(oVhdlFile)

    oRules.clear_violations()
    oRules.check_rules(
        bAllPhases=commandLineArguments.all_phases,
        lSkipPhase=commandLineArguments.skip_phase,
    )
    sOutputStd, sOutputErr = oRules.report_violations(
        commandLineArguments.output_format
    )
    fExitStatus = oRules.violations

    if commandLineArguments.junit:
        testCase = oRules.extract_junit_testcase(sFileName)
    else:
        testCase = None

    if commandLineArguments.json or commandLineArguments.quality_report:
        dJsonEntry["file_path"] = sFileName
        dJsonEntry["violations"] = oRules.extract_violation_dictionary()["violations"]

    return fExitStatus, testCase, dJsonEntry, sOutputStd, sOutputErr


def write_vhdl_file(oVhdlFile):
    try:
        with open(oVhdlFile.filename, 'w', encoding='utf-8') as oFile:
            for sLine in oVhdlFile.get_lines()[1:]:
                oFile.write(sLine + '\n')
    except PermissionError as err:
        print (err, "Could not write fixes back to file.")


def create_junit_testcase(sVhdlFileName, oException):
    '''
    Creates JUnit XML file listing all violations found.

    Parameters:

      sVhdlFileName (string)

    Returns: (junit testcase object)
    '''
    oTestcase = junit.testcase(sVhdlFileName, str(0), 'failure')
    oFailure = junit.failure('Failure')
    oFailure.add_text(oException.message)
    oTestcase.add_failure(oFailure)

    return oTestcase




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.process_statement.end_keyword, token.process_statement.end_postponed_keyword])
lTokens.append([token.process_statement.end_keyword, token.process_statement.end_process_keyword])


class rule_007(Rule):
    '''
    This rule checks for a single space after the **end** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end   process proc_a;

    **Fix**

    .. code-block:: vhdl

       end process proc_a;
    '''
    def __init__(self):
        Rule.__init__(self, 'process', '007', lTokens)




from vsg import parser
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils


class move_token_sequences_left_of_token(structure.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of parser object types
       object type to split a line at
    '''

    def __init__(self, name, identifier, lSequences, oLeftToken):
        structure.Rule.__init__(self, name=name, identifier=identifier)
        self.lSequences = lSequences
        self.oLeftToken = oLeftToken

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        lPrevious = []
        for lSequence in self.lSequences:
            if not lSequence[0] in lPrevious:
                aToi = oFile.get_tokens_bounded_by(lSequence[0], self.oLeftToken, bIncludeTillBeginningOfLine=True)
                lToi = utils.combine_two_token_class_lists(lToi, aToi)
            lPrevious.append(lSequence[0])
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            for iToken, oToken in enumerate(lTokens):
                for lSequence in self.lSequences:
                    bFound = False
                    if isinstance(oToken, lSequence[0]):
                        if utils.are_next_consecutive_token_types(lSequence, iToken, lTokens):
                            bFound = True
                            break
                        if utils.are_next_consecutive_token_types(lSequence[:-1], iToken, lTokens):
                            dAction = {}
                            dAction['num_tokens'] = len(lSequence) - 1
                        elif utils.are_next_consecutive_token_types(lSequence[:-2], iToken, lTokens):
                            dAction = {}
                            dAction['num_tokens'] = len(lSequence) - 2

                if bFound:
                    break
            else:
                sSolution = self.solution
                oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
                oViolation.set_action(dAction)
                oViolation.set_remap()
                oViolation.fix_blank_lines = True
                self.add_violation(oViolation)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        bInsertBlankLine = False
        if isinstance(lTokens[0], parser.whitespace):
            lTokens = lTokens[1:]
            bInsertBlankLine = True
        lMoveTokens = lTokens[0:dAction['num_tokens']]
        lTokens = lTokens[dAction['num_tokens']:]
        lTokens = lTokens[:-1] + lMoveTokens + [parser.whitespace(' ')] + [lTokens[-1]]
        lTokens = utils.remove_consecutive_whitespace_tokens(lTokens)
        if bInsertBlankLine:
            rules_utils.insert_blank_line(lTokens, 0)
        oViolation.set_tokens(lTokens)




from vsg.vhdlFile.classify import interface_function_specification
from vsg.vhdlFile.classify import interface_procedure_specification


def detect(iToken, lObjects):
    '''
        interface_subprogram_specification ::=
            interface_procedure_specification
          | interface_function_specification
    '''

    iCurrent = interface_procedure_specification.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = interface_function_specification.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




from vsg.token import waveform_element as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import expression


def classify_until(lUntils, iToken, lObjects):
    '''
    waveform_element ::=
        *value*_expression [ after *time*_expression ]
      | null [ after *time*_expression ]
    '''

    if utils.is_next_token('null', iToken, lObjects):
        iCurrent = utils.assign_next_token_required('null', token.null_keyword, iToken, lObjects)
    else:
        lMyUntils = lUntils
        lMyUntils.append('after')
        iCurrent = expression.classify_until(lMyUntils, iToken, lObjects)

    if utils.is_next_token('after', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('after', token.after_keyword, iCurrent, lObjects)
        iCurrent = expression.classify_until(lUntils, iCurrent, lObjects)

    return iCurrent




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.element_declaration.colon)

oStartToken = token.record_type_definition.record_keyword

oEndToken = token.record_type_definition.end_keyword


class rule_400(align_tokens_in_region_between_tokens):
    '''
    This rule checks the colons are in the same column for all elements in the block declarative part.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       type t_some_record is record
         element_1 : natural;
         some_other_element : natural;
         yet_another_element : natural;
       end record;

    **Fix**

    .. code-block:: vhdl

       type t_some_record is record
         element_1           : natural;
         some_other_element  : natural;
         yet_another_element : natural;
       end record;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'type', '400', lAlign, oStartToken, oEndToken)




from vsg.rules.whitespace_between_tokens import Rule

from vsg.token import case_statement as token


class rule_006(Rule):
    '''
    This rule checks for a single space between the **end** and **case** keywords.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

      case data is

      end    case;

    **Fix**

    .. code-block:: vhdl

      case data is

      end case;
    '''
    def __init__(self):
        Rule.__init__(self, 'case', '006')
        self.left_token = token.end_keyword
        self.right_token = token.end_case_keyword




from vsg import token

from vsg.rules import token_prefix

lTokens = []
lTokens.append(token.block_statement.block_label)
lTokens.append(token.block_statement.end_block_label)


class rule_601(token_prefix):
    '''
    This rule checks for valid prefixes on block labels.
    The default prefix is *blk_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       block_label : block is

    **Fix**

    .. code-block:: vhdl

       blk_block_label : block is
    '''

    def __init__(self):
        token_prefix.__init__(self, 'block', '601', lTokens)
        self.prefixes = ['blk_']
        self.solution = 'block label'




from vsg import parser
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment


class align_consecutive_lines_after_line_starting_with_token_and_stopping_with_token(alignment.Rule):
    '''
    '''

    def __init__(self, name, identifier):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.alignment = 'report'
        self.configuration.append('alignment')
        self.phase = 4
        self.lStartTokens = []
        self.lEndTokens = []
        self.indentAdjust = 1

    def _get_tokens_of_interest(self, oFile):
        lReturn = []
        lToi = oFile.get_tokens_starting_with_token_and_ending_with_one_of_possible_tokens(self.lStartTokens, self.lEndTokens, True, False, True)
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            lIndexes = get_index_after_carriage_return(lTokens)

            for iIndex in lIndexes:
                myToi = oToi.extract_tokens(iIndex, iIndex)
                myToi.set_meta_data('iSpaces', self._calculate_column(oFile, oToi, lTokens))
                myToi.set_meta_data('sWhitespace', self._expected_whitespace(oFile, oToi, lTokens))
                lReturn.append(myToi)

        return lReturn

    def _analyze(self, lToi):
        for oToi in lToi:
            self._check_leading_whitespace(oToi)

    def _check_leading_whitespace(self, oToi):
        sWhitespace = oToi.get_meta_data('sWhitespace')
        oToken = oToi.get_tokens()[0]
        if isinstance(oToken, parser.whitespace):
            if oToken.get_value() != sWhitespace:
                self._create_violation(oToi, 'adjust')
        else:
            self._create_violation(oToi, 'insert')

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()

        if dAction['action'] == 'adjust':
            lTokens[0].set_value(dAction['whitespace'])
        else:
            rules_utils.insert_new_whitespace(lTokens, 0, dAction['whitespace'])

        oViolation.set_tokens(lTokens)

    def _create_violation(self, oToi, sAction):
        sSolution = alignment_utils.build_solution(oToi.get_meta_data('sWhitespace'))
        oViolation = violation.New(oToi.get_line_number(), oToi, sSolution)
        dAction = {}
        dAction['action'] = sAction
        dAction['column'] = oToi.get_meta_data('iSpaces')
        dAction['whitespace'] = oToi.get_meta_data('sWhitespace')
        oViolation.set_action(dAction)
        self.add_violation(oViolation)

    def _calculate_column(self, oFile, oToi, lTokens):
        if self.alignment == 'report':
            iSpaces = oFile.get_column_of_token_index(oToi.get_start_index()) + 7
        else:
            iSpaces = (lTokens[0].indent + self.indentAdjust) * self.indentSize
        return iSpaces

    def _expected_whitespace(self, oFile, oToi, lTokens):
        if self.indentStyle == 'smart_tabs':
            if self.alignment == 'report':
                return (lTokens[0].indent) * '\t' + ' ' * (len(lTokens[0].get_value()) + 1)
            else:
                return (lTokens[0].indent + self.indentAdjust) * '\t'
        else:
            if self.alignment == 'report':
                iSpaces = oFile.get_column_of_token_index(oToi.get_start_index()) + (len(lTokens[0].get_value()) + 1)
            else:
                iSpaces = (lTokens[0].indent + self.indentAdjust) * self.indentSize
            return iSpaces * ' '


def get_index_after_carriage_return(lTokens):
    lReturn = []
    for iToken, oToken in enumerate(lTokens):
        if isinstance(oToken, parser.carriage_return):
            lReturn.append(iToken + 1)
    return lReturn





from vsg import parser

from vsg.rules.whitespace_between_tokens import Rule as WhitespaceRule

from vsg.vhdlFile import utils


class Rule(WhitespaceRule):
    '''
    Checks for a single space between two tokens.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    lTokens : list of token type pairs
       The tokens to check for a single space between
    '''

    def __init__(self, name, identifier, lTokens, oStart, oEnd):
        WhitespaceRule.__init__(self, name=name, identifier=identifier)
        self.lTokens = lTokens
        self.oStart = oStart
        self.oEnd = oEnd

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokens:
            lToi_a = oFile.get_sequence_of_tokens_matching_bounded_by_tokens([lTokenPair[0], parser.whitespace, lTokenPair[1]], self.oStart, self.oEnd)
            lToi = utils.combine_two_token_class_lists(lToi, lToi_a)
            lToi_a = oFile.get_sequence_of_tokens_matching_bounded_by_tokens(lTokenPair, self.oStart, self.oEnd)
            lToi = utils.combine_two_token_class_lists(lToi, lToi_a)
        return lToi




from vsg.rules import token_prefix

from vsg import token

lTokens = []
lTokens.append(token.process_statement.process_label)
lTokens.append(token.process_statement.end_process_label)


class rule_036(token_prefix):
    '''
    This rule checks for valid prefixes on process labels.
    The default prefix is *proc\_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       main: process () is

    **Fix**

    .. code-block:: vhdl

       proc_main: process () is
    '''

    def __init__(self):
        token_prefix.__init__(self, 'process', '036', lTokens)
        self.prefixes = ['proc_']
        self.solution = 'Process labels'




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.context_declaration.end_keyword)


class rule_020(token_indent):
    '''
    This rule checks the indent of the **end** keyword.

    **Violation**

    .. code-block:: vhdl

       context c1 is
          end context c1;

    **Fix**

    .. code-block:: vhdl

       context c1 is
       end context c1;
    '''

    def __init__(self):
        token_indent.__init__(self, 'context', '020', lTokens)




from vsg.vhdlFile.extract import tokens


def get_sequence_of_tokens_matching_bounded_by_tokens(lTokens, oStart, oEnd, lAllTokens, oTokenMap):

    lReturn = []

    lStart = oTokenMap.get_token_indexes(oStart)
    lEnd = oTokenMap.get_token_indexes(oEnd)
    lIndexes = []
    for iStart, iEnd in zip(lStart, lEnd):
        lIndexes.extend(oTokenMap.get_token_indexes_between_indexes(lTokens[0], iStart, iEnd))

    lIndexes.sort()

    for iIndex in lIndexes:
        iLine = oTokenMap.get_line_number_of_index(iIndex)
        for iToken, oToken in enumerate(lTokens):
            if not isinstance(lAllTokens[iToken + iIndex], oToken):
                break
        else:
            lReturn.append(tokens.New(iIndex, iLine, lAllTokens[iIndex:iIndex + len(lTokens)]))

    return lReturn




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.port_clause.port_keyword, token.port_clause.open_parenthesis])


class rule_003(Rule):
    '''
    This rule checks for a single space after the **port** keyword and (.

    **Violation**

    .. code-block:: vhdl

       port   (

       port(

    **Fix**

    .. code-block:: vhdl

       port (

       port (
    '''
    def __init__(self):
        Rule.__init__(self, 'port', '003', lTokens)
        self.solution = 'Change spacing between "port" and "(" to one space.'




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.logical_name_list.logical_name)


class rule_500(token_case):
    '''
    This rule checks the logical_name in a library_clause has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       library IEEE;

       library FIFO_dsn;

    **Fix**

    .. code-block:: vhdl

       library ieee;

       library fifo_dsn;
    '''

    def __init__(self):
        token_case.__init__(self, 'library', '500', lTokens)
        self.groups.append('case::name')
        self.configuration.append('case_exceptions')




from vsg.rules import existence_of_tokens_which_should_not_occur

from vsg import token

lTokens = []
lTokens.append(token.signal_declaration.assignment_operator)


class rule_007(existence_of_tokens_which_should_not_occur):
    '''
    This rule checks for default assignments in signal declarations.

    .. NOTE:: This rule requires the user to remove the default assignments.

    **Violation**

    .. code-block:: vhdl

       signal wr_en : std_logic := '0';

    **Fix**

    .. code-block:: vhdl

       signal wr_en : std_logic;
    '''

    def __init__(self):
        existence_of_tokens_which_should_not_occur.__init__(self, 'signal', '007', lTokens)
        self.solution = 'Remove default assignment.'



from vsg import parser


class is_keyword(parser.keyword):
    '''
    unique_id = subprogram_body : is_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class begin_keyword(parser.keyword):
    '''
    unique_id = subprogram_body : begin_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class end_keyword(parser.keyword):
    '''
    unique_id = subprogram_body : end_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class designator(parser.designator):
    '''
    unique_id = subprogram_body : designator
    '''

    def __init__(self, sString):
        parser.designator.__init__(self, sString)


class semicolon(parser.semicolon):
    '''
    unique_id = subprogram_body : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg import token

from vsg.rules import move_token_left_to_next_non_whitespace_token as Rule

oToken = token.full_type_declaration.semicolon


class rule_007(Rule):
    '''
    This rule checks the semicolon is on the same line as the **record** keyword.

    **Violation**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record
       ;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record t_record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '007', oToken)
        self.bInsertWhitespace = False
        self.subphase = 3





from vsg import parser


def classify(lTokens, lObjects):
    '''
    Classifies whitespace objects.
    '''
    for iToken, sToken in enumerate(lTokens):
        if string_contains_space(sToken):
            if is_string_literal(sToken):
                pass
            elif is_character_literal(sToken):
                pass
            else:
                lObjects[iToken] = parser.whitespace(sToken)
                if string_contains_tab(sToken):
                    lObjects[iToken].has_tab = True
        elif string_contains_tab(sToken):
            lObjects[iToken] = parser.whitespace(sToken)
            lObjects[iToken].has_tab = True


def string_contains_space(sToken):
    if ' ' in sToken:
        return True
    return False


def string_contains_tab(sToken):
    if '\t' in sToken:
        return True
    return False


def is_string_literal(sToken):
    if sToken[0] == '"' and sToken[-1] == '"':
        return True
    return False


def is_character_literal(sToken):
    if sToken[0] == "'" and sToken[-1] == "'":
        return True
    return False




from vsg.vhdlFile.classify import concurrent_assertion_statement
from vsg.vhdlFile.classify import concurrent_procedure_call_statement
from vsg.vhdlFile.classify import process_statement


def detect(iToken, lObjects):
    '''
    entity_statement ::=
        concurrent_assertion_statement
      | *passive*_concurrent_procedure_call_statement
      | *passive*_process_statement
      | *PSL*_PSL_Directive
    '''

    iCurrent = process_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = concurrent_assertion_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = concurrent_procedure_call_statement.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




import bisect

from vsg.vhdlFile.extract import tokens


def get_tokens_between_tokens_inclusive_while_storing_value_from_token(left_token, right_token, value_token, lAllTokens, oTokenMap):

    lReturn = []

    lStart, lEnd = oTokenMap.get_token_pair_indexes(left_token, right_token)

    lValueIndexes = oTokenMap.get_token_indexes(value_token)

    lValues = []
    lValuesPopped = []
    for iStart, iEnd in zip(lStart, lEnd):

        update_value_list(lValueIndexes, lValues, lValuesPopped, iStart)

        iLine = oTokenMap.get_line_number_of_index(iStart)

        oTokens = tokens.New(iStart, iLine, lAllTokens[iStart:iEnd + 1])
        sValue = get_matching_token_value(lValues, lAllTokens)
        oTokens.set_token_value(sValue)
        lReturn.append(oTokens)
        update_popped_list(lValuesPopped, lValues)

    return lReturn


def update_value_list(lValueIndexes, lValues, lValuesPopped, iLeftIndex):
    for iValue in lValueIndexes:
        if iLeftIndex > iValue and iValue not in lValuesPopped and iValue not in lValues:
            lValues.append(iValue)


def get_matching_token_value(lValues, lAllTokens):
    try:
        oValueToken = lAllTokens[lValues[-1]]
        return oValueToken.get_value()
    except IndexError:
        return None


def update_popped_list(lValuesPopped, lValues):
    try:
        lValuesPopped.append(lValues.pop(-1))
    except IndexError:
        pass




from vsg.rules import multiline_simple_structure as Rule

from vsg import token

lTokenPairs = []
lTokenPairs.append([token.simple_variable_assignment.assignment, token.simple_variable_assignment.semicolon])
lTokenPairs.append([token.conditional_variable_assignment.assignment, token.conditional_variable_assignment.semicolon])


class rule_007(Rule):
    '''
    This rule checks the structure of simple and conditional variable assignments.

    |configuring_simple_multiline_structure_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en :=
         '0' when q_wr_en = '1' else
                '1';

       w_foo :=
         I_FOO when ((I_BAR = '1') and
                            (I_CRUFT = '1')) else
                '0';

    **Fix**

    .. code-block:: vhdl

       wr_en := '0' when q_wr_en = '1' else
                '1';

       w_foo := I_FOO when ((I_BAR = '1') and
                            (I_CRUFT = '1')) else
                '0';
    '''

    def __init__(self):
        Rule.__init__(self, 'variable_assignment', '007', lTokenPairs)




from vsg import parser
from vsg import token

from vsg.rules import consistent_token_case

lTokens = []
lTokens.append(token.procedure_specification.designator)

lIgnore = []
lIgnore.append(parser.whitespace)
lIgnore.append(parser.carriage_return)
lIgnore.append(parser.blank_line)


class rule_507(consistent_token_case):
    '''
    This rule checks for consistent capitalization of procedure names.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

         procedure average_samples is
         begin
         end procedure average_samples

       begin

         Average_samples;

         PROC1 : process () is
         begin

            AVERAGE_SAMPLES;

         end process;

       end architecture rtl;

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         procedure average_samples is
         begin
         end procedure average_samples

       begin

         average_samples;

         PROC1 : process () is
         begin

            average_samples;

         end process;

       end architecture rtl;
    '''

    def __init__(self):
        consistent_token_case.__init__(self, 'procedure', '507', lTokens, lIgnore)




from vsg.rules import move_token as Rule

from vsg import token

oToken = token.port_clause.close_parenthesis


class rule_014(Rule):
    '''
    This rule checks the location of the closing ")" character for the port clause.

    The default location is on a line by itself.

    |configuring_move_token_rules_link|

    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic);

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   std_logic;
         DATA     : inout std_logic
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'port', '014', oToken)
        self.preserve_comment = True




from vsg.token import entity_specification as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import entity_name_list


def classify(iToken, lObjects):
    '''
    entity_specification ::=
        entity_name_list : entity_class
    '''

    iCurrent = entity_name_list.classify(iToken, lObjects)

    iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)

    iCurrent = utils.assign_next_token(token.entity_class, iCurrent, lObjects)

    return iCurrent




from vsg import violation


def check_for_case_violation(oToi, self, check_prefix=False, check_suffix=False, check_whole=False, iIndex=0, iLine=None):
    sObjectValue = get_token_value(oToi, iIndex)
    iMyLine = get_violation_line(oToi, iLine)
    oViolation = None

    if is_lower_case_without_prefix_or_suffix_or_whole_exception(self.case, check_prefix, check_suffix, check_whole):
        oViolation = check_for_lower_case(sObjectValue, oToi, iIndex, iMyLine)

    if is_lower_case_with_prefix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_lower_case_with_prefix_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_lower_case_with_suffix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_lower_case_with_suffix_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_lower_case_with_prefix_and_suffix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_lower_case_with_prefix_and_suffix_exceptions(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_upper_case_without_prefix_or_suffix_or_whole_exception(self.case, check_prefix, check_suffix, check_whole):
        oViolation = check_for_upper_case(sObjectValue, oToi, iIndex, iMyLine)

    if is_upper_case_with_prefix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_upper_case_with_prefix_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_upper_case_with_suffix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_upper_case_with_suffix_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_upper_case_with_prefix_and_suffix_exception(self.case, check_prefix, check_suffix):
        oViolation = check_for_upper_case_with_prefix_and_suffix_exceptions(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_lower_case_with_whole_exception(self.case, check_whole):
        oViolation = check_for_lower_case_with_whole_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    if is_upper_case_with_whole_exception(self.case, check_whole):
        oViolation = check_for_upper_case_with_whole_exception(sObjectValue, self, oToi, iIndex, iMyLine)

    return oViolation


def get_token_value(oToi, iIndex):
    return oToi.get_tokens()[iIndex].get_value()


def prefix_detected(sString, lPrefixes):
    sLowerString = sString.lower()
    for sPrefix in lPrefixes:
       if sLowerString.startswith(sPrefix.lower()):
           return True
    return False


def get_matched_prefix(sString, lPrefixes):
    sLowerString = sString.lower()
    for sPrefix in lPrefixes:
       if sLowerString.startswith(sPrefix.lower()):
           return sPrefix


def extract_prefix(sString, sPrefix):
    return sString[0:len(sPrefix)]


def remove_prefix(sString, sPrefix):
    return sString[len(sPrefix):]


def suffix_detected(sString, lSuffixes):
    sLowerString = sString.lower()
    for sSuffix in lSuffixes:
       if sLowerString.endswith(sSuffix.lower()):
           return True
    return False


def get_matched_suffix(sString, lSuffixes):
    sLowerString = sString.lower()
    for sSuffix in lSuffixes:
       if sLowerString.endswith(sSuffix.lower()):
           return sSuffix


def extract_suffix(sString, sSuffix):
    return sString[len(sString) - len(sSuffix):]


def remove_suffix(sString, sSuffix):
    return sString[0:len(sString) - len(sSuffix)]


def create_case_violation(sActual, sExpected, oToi, iIndex, iLine):
    sSolution = 'Change "' + sActual + '" to "' + sExpected + '"'
    oViolation = violation.New(iLine, oToi, sSolution)
    dAction = {}
    dAction['value'] = sExpected
    dAction['index'] = iIndex
    oViolation.set_action(dAction)
    return oViolation


def get_violation_line(oToi, iLine):
    if iLine is None:
        return oToi.get_line_number()
    return iLine


def is_exception_enabled(lList):
    if len(lList) == 0:
        return False
    return True


def is_lower_case_without_prefix_or_suffix_or_whole_exception(sCase, bPrefix, bSuffix, bWhole):
    if sCase == 'lower' and not bPrefix and not bSuffix and not bWhole:
        return True
    return False


def is_lower_case_with_prefix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'lower' and bPrefix and not bSuffix:
        return True
    return False


def is_lower_case_with_whole_exception(sCase, bWhole):
    if sCase == 'lower' and bWhole:
        return True
    return False


def is_lower_case_with_suffix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'lower' and not bPrefix and bSuffix:
        return True
    return False


def is_lower_case_with_prefix_and_suffix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'lower' and bPrefix and bSuffix:
        return True
    return False


def is_upper_case_without_prefix_or_suffix_or_whole_exception(sCase, bPrefix, bSuffix, bWhole):
    if sCase == 'upper' and not bPrefix and not bSuffix and not bWhole:
        return True
    return False


def is_upper_case_with_prefix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'upper' and bPrefix and not bSuffix:
        return True
    return False


def is_upper_case_with_suffix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'upper' and not bPrefix and bSuffix:
        return True
    return False


def is_upper_case_with_prefix_and_suffix_exception(sCase, bPrefix, bSuffix):
    if sCase == 'upper' and bPrefix and bSuffix:
        return True
    return False


def is_upper_case_with_whole_exception(sCase, bWhole):
    if sCase == 'upper' and bWhole:
        return True
    return False


def check_for_lower_case(sObjectValue, oToi, iIndex, iLine):
    if not sObjectValue.islower():
        return create_case_violation(sObjectValue, sObjectValue.lower(), oToi, iIndex, iLine)


def check_for_lower_case_with_whole_exception(sObjectValue, self, oToi, iIndex, iLine):
    if exception_found(sObjectValue, self):
        return check_for_exception(sObjectValue, self, oToi, iIndex, iLine)
    elif not sObjectValue.islower():
        return create_case_violation(sObjectValue, sObjectValue.lower(), oToi, iIndex, iLine)


def exception_found(sObjectValue, self):
    if sObjectValue.lower() in self.case_exceptions_lower:
        return True
    return False


def check_for_exception(sObjectValue, self, oToi, iIndex, iLine):
    iIndex = self.case_exceptions_lower.index(sObjectValue.lower())
    if sObjectValue != self.case_exceptions[iIndex]:
        return create_case_violation(sObjectValue, self.case_exceptions[iIndex], oToi, iIndex, iLine)


def check_for_upper_case(sObjectValue, oToi, iIndex, iLine):
    if not sObjectValue.isupper():
        return create_case_violation(sObjectValue, sObjectValue.upper(), oToi, iIndex, iLine)


def check_for_upper_case_with_whole_exception(sObjectValue, self, oToi, iIndex, iLine):
    if exception_found(sObjectValue, self):
        return check_for_exception(sObjectValue, self, oToi, iIndex, iLine)
    elif not sObjectValue.isupper():
        return create_case_violation(sObjectValue, sObjectValue.upper(), oToi, iIndex, iLine)


def check_for_lower_case_with_prefix_exception(sObjectValue, self, oToi, iIndex, iLine):
    if prefix_detected(sObjectValue, self.prefix_exceptions):
        sDesiredPrefix = get_matched_prefix(sObjectValue, self.prefix_exceptions)
        sActualPrefix = extract_prefix(sObjectValue, sDesiredPrefix)
        sConstant = remove_prefix(sObjectValue, sActualPrefix)
        sExpected = sDesiredPrefix + sConstant.lower()
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_lower_case(sObjectValue, oToi, iIndex, iLine)


def check_for_upper_case_with_prefix_exception(sObjectValue, self, oToi, iIndex, iLine):
    if prefix_detected(sObjectValue, self.prefix_exceptions):
        sDesiredPrefix = get_matched_prefix(sObjectValue, self.prefix_exceptions)
        sActualPrefix = extract_prefix(sObjectValue, sDesiredPrefix)
        sConstant = remove_prefix(sObjectValue, sActualPrefix)
        sExpected = sDesiredPrefix + sConstant.upper()
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_upper_case(sObjectValue, oToi, iIndex, iLine)


def check_for_lower_case_with_suffix_exception(sObjectValue, self, oToi, iIndex, iLine):
    if suffix_detected(sObjectValue, self.suffix_exceptions):
        sDesiredSuffix = get_matched_suffix(sObjectValue, self.suffix_exceptions)
        sActualSuffix = extract_suffix(sObjectValue, sDesiredSuffix)
        sConstant = remove_suffix(sObjectValue, sActualSuffix)
        sExpected = sConstant.lower() + sDesiredSuffix
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_lower_case(sObjectValue, oToi, iIndex, iLine)


def check_for_upper_case_with_suffix_exception(sObjectValue, self, oToi, iIndex, iLine):
    if suffix_detected(sObjectValue, self.suffix_exceptions):
        sDesiredSuffix = get_matched_suffix(sObjectValue, self.suffix_exceptions)
        sActualSuffix = extract_suffix(sObjectValue, sDesiredSuffix)
        sConstant = remove_suffix(sObjectValue, sActualSuffix)
        sExpected = sConstant.upper() + sDesiredSuffix
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_upper_case(sObjectValue, oToi, iIndex, iLine)


def check_for_lower_case_with_prefix_and_suffix_exceptions(sObjectValue, self, oToi, iIndex, iLine):
    if prefix_detected(sObjectValue, self.prefix_exceptions) and suffix_detected(sObjectValue, self.suffix_exceptions):
        sDesiredPrefix = get_matched_prefix(sObjectValue, self.prefix_exceptions)
        sActualPrefix = extract_prefix(sObjectValue, sDesiredPrefix)
        sConstant = remove_prefix(sObjectValue, sActualPrefix)
        sDesiredSuffix = get_matched_suffix(sConstant, self.suffix_exceptions)
        sActualSuffix = extract_suffix(sConstant, sDesiredSuffix)
        sConstant = remove_suffix(sConstant, sActualSuffix)
        sExpected = sDesiredPrefix + sConstant.lower() + sDesiredSuffix
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_lower_case(sObjectValue, oToi, iIndex, iLine)


def check_for_upper_case_with_prefix_and_suffix_exceptions(sObjectValue, self, oToi, iIndex, iLine):
    if prefix_detected(sObjectValue, self.prefix_exceptions) and suffix_detected(sObjectValue, self.suffix_exceptions):
        sDesiredPrefix = get_matched_prefix(sObjectValue, self.prefix_exceptions)
        sActualPrefix = extract_prefix(sObjectValue, sDesiredPrefix)
        sConstant = remove_prefix(sObjectValue, sActualPrefix)
        sDesiredSuffix = get_matched_suffix(sConstant, self.suffix_exceptions)
        sActualSuffix = extract_suffix(sConstant, sDesiredSuffix)
        sConstant = remove_suffix(sConstant, sActualSuffix)
        sExpected = sDesiredPrefix + sConstant.upper() + sDesiredSuffix
        if sObjectValue != sExpected:
            return create_case_violation(sObjectValue, sExpected, oToi, iIndex, iLine)
    else:
        return check_for_upper_case(sObjectValue, oToi, iIndex, iLine)




from vsg.vhdlFile.classify import interface_incomplete_type_declaration


def detect(iToken, lObjects):
    '''
    interface_type_declaration ::=
        interface_incomplete_type_declaration
    '''

    return interface_incomplete_type_declaration.detect(iToken, lObjects)




from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import entity_declarative_part
from vsg.vhdlFile.classify import entity_header
from vsg.vhdlFile.classify import entity_statement_part

from vsg.token import entity_declaration as token


def detect(iToken, lObjects):
    '''
    entity_declaration ::=
        entity identifier is
            entity_header
            entity_declarative_part
        [ begin
            entity_statement_part ]
        end [ entity ] [ entity_simple_name ] ;
    '''

    if utils.is_next_token('entity', iToken, lObjects):
        return classify(iToken, lObjects)
    else:
        return iToken


def classify(iToken, lObjects):
    iCurrent = classify_opening_declaration(iToken, lObjects)

    iCurrent = entity_header.detect(iCurrent, lObjects)

    iCurrent = entity_declarative_part.detect(iCurrent, lObjects)

    if utils.is_next_token('begin', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('begin', token.begin_keyword, iCurrent, lObjects)
        iCurrent = entity_statement_part.detect(iCurrent, lObjects)

    iCurrent = classify_closing_declaration(iCurrent, lObjects)

    return iCurrent


def classify_opening_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('entity', token.entity_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token(token.identifier, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    return iCurrent


def classify_closing_declaration(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('end', token.end_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('entity', token.end_entity_keyword, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_if_not(';', token.entity_simple_name, iCurrent, lObjects)
    iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.logical_operator.and_operator)
lTokens.append(token.logical_operator.or_operator)
lTokens.append(token.logical_operator.nand_operator)
lTokens.append(token.logical_operator.nor_operator)
lTokens.append(token.logical_operator.xor_operator)
lTokens.append(token.logical_operator.xnor_operator)


class rule_500(token_case):
    '''
    This rule checks logical operators have proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       a <= b AND c;

    **Fix**

    .. code-block:: vhdl

       a <= b and c;
    '''

    def __init__(self):
        token_case.__init__(self, 'logical_operator', '500', lTokens)
        self.groups.append('case::keyword')




import os
import importlib
import inspect

from vsg import exceptions

from . import deprecated_rule
from . import junit
from . import report
from . import utils
from . import severity


def get_python_modules_from_directory(sDirectoryName, lModules):
    '''
    Returns a list of files with an extension of py from a directory.
    It ignores files starting with a double underscore __.

    Parameters:

      sDirectoryName (string)

    Modifies:

      lModules (string list)
    '''
    lDirectoryContents = os.listdir(sDirectoryName)
    for sFileName in lDirectoryContents:
        if sFileName.endswith('.py') and not sFileName.startswith('__'):
            lModules.append(sFileName.replace('.py', ''))


def get_rules_from_module(lModules, lRules):
    '''
    Returns a list of files that start with "rule_".

    Parameters:

      lModules (list)

    Modifies:

      lRules (object list)
    '''
    for sModuleName in lModules:
        for name, obj in inspect.getmembers(importlib.import_module(sModuleName)):
            if name.startswith('rule_') and name != 'rule_item':
                lRules.append(obj())


def load_local_rules(sDirectoryName):
    '''
    Loads rules from the directory passed to this routine.

    Parameters:

      sDirectoryName (string)

    Returns: (string list)
    '''
    lLocalModules = []
    get_python_modules_from_directory(sDirectoryName, lLocalModules)
    lRules = []
    get_rules_from_module(lLocalModules, lRules)
    return lRules


def load_rules():
    '''
    Loads rules from the vsg/rules directory.

    Parameters:  None

    Returns:  (rule object list)
    '''
    lRules = []
    for name, oPackage in inspect.getmembers(importlib.import_module('vsg.rules')):
        if inspect.ismodule(oPackage):
            for name, oRule in inspect.getmembers(oPackage):
                if inspect.isclass(oRule) and name.startswith('rule_'):
                    lRules.append(oRule())

    return lRules


def maximum_phase(lRules):
    '''
    Determines the maximum phase number from all the rules.

    Parameters:
      lRules (rule object list)

    Returns: (integer)
    '''
    maximumPhaseNumber = 0
    for oRule in lRules:
        if oRule.phase > maximumPhaseNumber:
            maximumPhaseNumber = oRule.phase
    return maximumPhaseNumber


class rule_list():
    '''
    Contains a list of all rules to be checked.
    It loads all base rules.
    Localized rules are loaded if specified.

    Parameters:

      oVhdlFile: (vhdlFile object)

      oSeverityList: (severity list object)

      sLocalRulesDirectory: (string) (optional)
    '''
    def __init__(self, oVhdlFile, oSeverityList, sLocalRulesDirectory=None):
        self.rules = load_rules()
        if sLocalRulesDirectory:
            self.rules.extend(load_local_rules(sLocalRulesDirectory))
        self.iNumberRulesRan = 0
        self.lastPhaseRan = 0
        self.oVhdlFile = oVhdlFile
        self.maximumPhase = maximum_phase(self.rules)
        self.violations = False
        self.oSeverityList = oSeverityList

    def fix(self, iFixPhase=7, lSkipPhase=None, dFixOnly=None):
        '''
        Applies fixes to all violations found.

        Parameters:

          iFixPhase : (integer)

          lSkipPhases : (list of integers)

          dFixOnly : (fix list dictionary)
        '''
        if lSkipPhase is None:
            lSkipPhase = []
        for phase in range(1, int(iFixPhase) + 1):
            if phase in lSkipPhase:
                if phase == 1:
                    self.oVhdlFile.set_token_indent()
                continue

            # Update indents before checking indent
            if phase == 4:
                self.oVhdlFile.set_token_indent()

            for subphase in range(0, 5):
                lRules = self.get_rules_in_phase(phase)
                lRules = self.get_rules_in_subphase(lRules, subphase)
                lRules = filter_out_disabled_rules(lRules)
                for oRule in lRules:
                    if oRule.severity.type == severity.error_type:
                        oRule.fix(self.oVhdlFile, dFixOnly)
                    else:
                        oRule.analyze(self.oVhdlFile)

            if phase == 1:
                self.oVhdlFile.fix_blank_lines()
                self.oVhdlFile.fix_trailing_whitespace()
                self.oVhdlFile.update_token_map()

    def get_rules_in_phase(self, iPhaseNumber):
        '''
        Returns a list of rules in a given phase.

        Parameters:

          iPhaseNumber : (integer)

        Returns: (list of rule objects)
        '''
        lReturn = []
        for oRule in self.rules:
            if oRule.phase == iPhaseNumber:
                lReturn.append(oRule)
        return lReturn

    def get_rules_in_subphase(self, lRules, iSubPhase):
        '''
        Returns a list of rules in a given subphase.

        Parameters:

          lRules : (list of rule objects)

          iSubPhase : (integer)

        Returns: (list of rule objects)
        '''
        lReturn = []
        for oRule in lRules:
            if oRule.subphase == iSubPhase:
                lReturn.append(oRule)
        return lReturn

    def check_rules(self, bAllPhases=False, lSkipPhase=None):
        '''
        Analyzes all rules in increasing phase order.
        If there is a violation in a phase, analysis is halted.

        Parameters:

            bAllPhases : (boolean)
            lSkipPhase : (list of integers)
        '''
        if lSkipPhase is None:
            lSkipPhase = []
        self.iNumberRulesRan = 0
        iFailures = 0
        self.violations = False
        for phase in range(1, 8):
            if phase in lSkipPhase:
                continue

            for subphase in range(0, 5):
                lRules = self.get_rules_in_phase(phase)
                lRules = self.get_rules_in_subphase(lRules, subphase)
                lRules = filter_out_disabled_rules(lRules)

                for oRule in lRules:
                    oRule.analyze(self.oVhdlFile)
                    if oRule.severity.type == severity.error_type:
                        iFailures += len(oRule.violations)
                    self.iNumberRulesRan += 1
                self.lastPhaseRan = phase
                if iFailures > 0:
                    self.violations = True
            if self.violations:
                if not bAllPhases:
                    break

    def report_violations(self, sOutputFormat):
        '''
        Prints out violations to stdout.

        Parameters:

          sOutputFormat (string)
        '''
        dRunInfo = {}
        dRunInfo['filename'] = self.oVhdlFile.filename
        dRunInfo['stopPhase'] = 7
        dRunInfo['violations'] = []

        for oRule in self.rules:
            if oRule.has_violations():
                lViolations = oRule.get_violations()
                dRunInfo['violations'].extend(lViolations)
        dRunInfo['violations'] = sorted(dRunInfo['violations'], key=lambda x: int(x['lineNumber']))
        dRunInfo['stopPhase'] = self.lastPhaseRan
        dRunInfo['num_rules_checked'] = self.get_number_of_rules_ran()
        dRunInfo['total_violations'] = len(dRunInfo['violations'])
        dRunInfo['maxSeverityNameLength'] = self.oSeverityList.iMaxNameLength
        dRunInfo['severities'] = {}

        for oSeverity in self.oSeverityList.get_severities():
            dRunInfo['severities'][oSeverity.name] = 0
        for dViolation in dRunInfo['violations']:
            name = dViolation['severity']['name']
            dRunInfo['severities'][name] = dRunInfo['severities'][name] + 1

        if sOutputFormat == 'vsg':
            sOutputStd, sOutputErr = report.vsg_stdout.print_output(dRunInfo)
        elif sOutputFormat == 'syntastic':
            sOutputStd, sOutputErr = report.syntastic_stdout.print_output(dRunInfo)
        else:
            sOutputStd, sOutputErr = report.summary_stdout.print_output(dRunInfo)
        return sOutputStd, sOutputErr

    def configure(self, oConfig):
        '''
        Configures individual rules based on dictionary passed.

        Parameters:

          configurationFile: (dictionary)
        '''
        lDeprecatedMessages = []
        configurationFile = oConfig.dConfig
        if configurationFile and 'rule' in configurationFile:
            self._validate_configuration_rule_exists(configurationFile)
            for oRule in self.rules:
                lDeprecatedMessages.extend(oRule.configure(oConfig))

        if len(lDeprecatedMessages) != 0:
            sErrorMessage = f'ERROR: Invalid configuration of file {self.oVhdlFile.filename}' + '\n'
            for sMessage in lDeprecatedMessages:
                sErrorMessage += sMessage + '\n'
            raise exceptions.ConfigurationError(sErrorMessage)

        try:
            if configurationFile['debug']:
                for oRule in self.rules:
                    oRule.set_debug()
        except KeyError:
            pass

    def _validate_configuration_rule_exists(self, configurationFile):
        '''
        Validates rules called out in the configuration files exist in the rule set.

        If a rule does not exist then:

          1) an error message will be printed
          2) tool will exit with a status of 1

        Parameters:

          configurationFile: (dictionary)

        Returns:  nothing
        '''
        lRuleNames = self.get_list_of_rule_names()
        for sRule in configurationFile['rule']:
            if rule_does_not_exist_in_list(sRule, lRuleNames):
                sErrorMessage = 'ERROR: Rule ' + sRule + ' referenced in configuration could not be found'
                raise exceptions.ConfigurationError(sErrorMessage)

    def get_list_of_rule_names(self):
        lReturn = []
        for oRule in self.rules:
            lReturn.append(oRule.get_unique_id())
        return lReturn

    def extract_junit_testcase(self, sVhdlFileName):
        '''
        Creates JUnit XML file listing all violations found.

        Parameters:

          sVhdlFileName (string)

        Returns: (junit testcase object)
        '''
        oTestcase = junit.testcase(sVhdlFileName, str(0), 'failure')
        oFailure = junit.failure('Failure')
        for oRule in self.rules:
            if len(oRule.violations) > 0 and oRule.severity.type == severity.error_type:
                for dViolation in oRule.violations:
                    sLine = oRule.name + '_' + oRule.identifier + ': '
                    sLine += str(utils.get_violation_line_number(dViolation)) + ' : '
                    sLine += dViolation.get_solution()
                    oFailure.add_text(sLine)
        oTestcase.add_failure(oFailure)

        return oTestcase

    def extract_violation_dictionary(self):
        '''
        Creates a dictionary of violations which were found.

        Returns: (dictionary)
        '''
        dReturn = {}
        dReturn['violations'] = []
        for oRule in self.rules:
            if oRule.has_violations:
                for oViolation in oRule.violations:
                    dTemp = {}
                    dTemp['rule'] = oRule.unique_id
                    dTemp['linenumber'] = oViolation.get_line_number()
                    dTemp['severity'] = oRule.severity.name
                    dTemp['solution'] = oViolation.get_solution()
                    dReturn['violations'].append(dTemp)
        return dReturn

    def get_configuration(self):
        '''
        Returns a dictionary with every rule and how it is configured.

        Parameters:

          None

        Returns: (dictionary)
        '''
        dConfiguration = {}
        for oRule in self.rules:
            if is_rule_deprecated(oRule):
                continue
            dConfiguration[oRule.unique_id] = oRule.get_configuration()
        return dConfiguration

    def clear_violations(self):
        for oRule in self.rules:
            oRule.clear_violations()

    def get_number_of_rules_ran(self):
        return self.iNumberRulesRan


def filter_out_disabled_rules(lRules):
    '''
    Removes rules which are disabled from a list of rule objects.

    Parameters:

      lRules : (list of rule objects)

    Returns: (list of rule objects)
    '''
    lReturn = []
    for oRule in lRules:
        if not oRule.disable:
            lReturn.append(oRule)
    return lReturn


def is_rule_deprecated(oRule):
    if isinstance(oRule, deprecated_rule.Rule):
        return True
    return False


def rule_does_not_exist_in_list(sRule, lRuleNames):
    if is_global_configuration(sRule):
        return False
    if is_group_configuration(sRule):
        return False
    if sRule not in lRuleNames:
        return True
    return False


def is_global_configuration(sName):
    if sName == 'global':
        return True
    return False


def is_group_configuration(sName):
    if sName == 'group':
        return True
    return False




from vsg.vhdlFile.classify import primary_unit
from vsg.vhdlFile.classify import secondary_unit


def detect(iToken, lObjects):
    '''
    library_unit ::=
        primary_unit
      | secondary_unit
    '''

    iCurrent = primary_unit.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    iCurrent = secondary_unit.detect(iToken, lObjects)
    if iCurrent != iToken:
        return iCurrent

    return iToken




from vsg import token

from vsg.rules.whitespace_between_token_pairs_bounded_by_tokens import Rule

lTokens = []
lTokens.append([token.association_element.assignment, token.association_element.actual_part])

lStart = token.port_map_aspect.open_parenthesis
lEnd = token.port_map_aspect.close_parenthesis


class rule_007(Rule):
    '''
    This rule checks for a single space after the **=>** operator in port maps.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       U_FIFO : FIFO
         port map (
           WR_EN    =>   wr_en,
           RD_EN    =>rd_en,
           OVERFLOW =>     overflow
         );

    **Fix**

    .. code-block:: vhdl

       U_FIFO : FIFO
         port map (
           WR_EN    => wr_en,
           RD_EN    => rd_en,
           OVERFLOW => overflow
         );
    '''
    def __init__(self):
        Rule.__init__(self, 'port_map', '007', lTokens, lStart, lEnd)




from vsg import parser

from vsg.vhdlFile.extract import tokens


def get_tokens_from_line(iLineNumber, lAllTokens, oTokenMap):
    lIndexes = oTokenMap.get_token_indexes(parser.carriage_return)
    iLine = iLineNumber - 2

    iStart = lIndexes[iLine] + 1
    iEnd = lIndexes[iLine + 1] + 1

    lTemp = lAllTokens[iStart:iEnd]

    return tokens.New(iStart, iLineNumber, lTemp)




from vsg import parser


class logical_name(parser.logical_name):
    '''
    unique_id = logical_name_list : logical_name
    '''

    def __init__(self, sString):
        parser.logical_name.__init__(self, sString)


class comma(parser.comma):
    '''
    unique_id = logical_name_list : comma
    '''

    def __init__(self, sString=','):
        parser.comma.__init__(self)




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.use_clause.keyword)


class rule_008(token_indent):
    '''
    This rule checks the indent of the **use** keyword.

    **Violation**

    .. code-block:: vhdl

       library ieee;
       use ieee.std_logic_1164.all;
            use ieee.std_logic_unsigned.all;

    **Fix**

    .. code-block:: vhdl

       library ieee;
         use ieee.std_logic_1164.all;
         use ieee.std_logic_unsigned.all;
    '''

    def __init__(self):
        token_indent.__init__(self, 'library', '008', lTokens)




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class multiline_array_alignment(alignment.Rule):
    '''
    This rule checks the alignment of multiline constants that contain arrays.

    |configuring_multiline_indent_rules_link|

    .. NOTE:: The structure of multiline array constants is handled by the rule `constant_016 <constant_rules.html#constant-016>`_.

    **Violation**

    .. code-block:: vhdl

       constant rom : romq_type :=
       (
                0,
            65535,
            32768
         );

    **Fix**

    .. code-block:: vhdl

       constant rom : romq_type :=
       (
         0,
         65535,
         32768
       );
    '''

    def __init__(self, name, identifier, lTokenPairs):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.subphase = 2
        self.lTokenPairs = lTokenPairs
        self.align_left = 'no'
        self.configuration.append('align_left')
        self.align_paren = 'yes'
        self.configuration.append('align_paren')

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            aToi = remove_non_arrays(aToi)
            populate_toi_parameters(aToi, oFile)
            aToi = remove_single_line_assignments(aToi)
            lToi = utils.combine_two_token_class_lists(lToi, aToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            oLines = lines(oToi)
            oLines.set_first_line_indent(oToi.iFirstLineIndent)
            self.calculate_expected_indents(oToi, oLines)
            check_indents(self, oToi, oLines)

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
        if dAction['action'] == 'adjust':
            lTokens[0].set_value(dAction['whitespace'])
        else:
            rules_utils.insert_new_whitespace(lTokens, 0, dAction['whitespace'])

        oViolation.set_tokens(lTokens)

    def calculate_expected_indents(self, oToi, oLines):
        dExpectedIndent = {}
        if not align_paren(self) and align_left(self):
            dExpectedIndent = self.analyze_align_left_true_align_paren_false(oToi, oLines)
        if align_paren(self) and not align_left(self):
            dExpectedIndent = self.analyze_align_left_false_align_paren_true(oToi, oLines)
        if align_paren(self) and align_left(self):
            dExpectedIndent = self.analyze_align_left_true_align_paren_true(oToi, oLines)

        return dExpectedIndent

    def analyze_align_left_true_align_paren_false(self, oToi, oLines):
        oLines.iIndent = oLines.get_first_line_indent()
        oLines.iParens = 0
        oLines.indentSize = self.indentSize
        for oLine in oLines.lLines:
            if oLine.isFirst:
                oLines.set_first_line_expected_indent(oLines.iIndent)
            else:
                check_left_aligned_line(oLine, oLines)

            oLines.iParens += oLine.get_delta_parens()

    def analyze_align_left_false_align_paren_true(self, oToi, oLines):

        for oLine in oLines.lLines:
            if oLine.isFirst:
                check_first_line(oLine, oLines, oToi, self.indentSize)
            elif oLine.isLast:
                check_last_line(oLine, oLines, self.indentSize)
            else:
                check_middle_line(oLine, oLines, self.indentSize)

    def analyze_align_left_true_align_paren_true(self, oToi, oLines):

        for oLine in oLines.lLines:
            if oLine.isFirst:
                check_my_first_line(oLine, oLines, oToi, self.indentSize)
            elif oLine.isLast:
                check_last_line(oLine, oLines, self.indentSize)
            else:
                check_middle_line(oLine, oLines, self.indentSize)


def _set_indent(iToken, lTokens):
    iReturn = 0
    if isinstance(lTokens[iToken + 1], parser.whitespace):
        iReturn = len(lTokens[iToken + 1].get_value())
    else:
        iReturn = 0
    return iReturn


def check_left_aligned_line(oLine, oLines):
    if rules_utils.token_list_begins_with_close_paren(oLine.tokens):
        oLine.set_expected_indent(oLines.iIndent + oLines.iParens * oLines.indentSize - oLines.indentSize)
    else:
        oLine.set_expected_indent(oLines.iIndent + oLines.iParens * oLines.indentSize)


def check_first_line(oLine, oLines, oToi, iIndentStep):
    iIndent = oLines.get_first_line_indent()
    oLine.set_expected_indent(iIndent)

    iAdjust = oToi.iAssignColumn
    oLines.update_parens(oLine, iIndentStep, iAdjust)

    if oLines.no_parens():
        oLines.iNextIndent = iIndent
    else:
        oLines.iNextIndent = oLines.lParens[-1].iExpectedColumn


def check_last_line(oLine, oLines, iIndentStep):
    set_expected_indent_of_line(oLine, oLines, iIndentStep)
    oLines.update_parens(oLine, iIndentStep)


def check_middle_line(oLine, oLines, iIndentStep):
    set_expected_indent_of_line(oLine, oLines, iIndentStep)
    oLines.update_parens(oLine, iIndentStep)
    oLines.iNextIndent = oLines.lParens[-1].iExpectedColumn


def set_expected_indent_of_line(oLine, oLines, iIndentStep):
    if oLine.starts_with_close_paren():
        oLine.set_expected_indent(oLines.iNextIndent - iIndentStep)
    else:
        oLine.set_expected_indent(oLines.iNextIndent)


def check_my_first_line(oLine, oLines, oToi, iIndentStep):
    iIndent = oLines.get_first_line_indent()
    oLine.set_expected_indent(iIndent)

    iAdjust = oToi.iAssignColumn
    oLines.update_parens(oLine, iIndentStep, iAdjust)

    if oLines.no_parens():
        oLines.iNextIndent = iIndent
    else:
        oLines.iNextIndent = len(oLines.lParens)*iIndentStep + iIndent

    for iParen, oParen in enumerate(oLine.parens):
        oParen.iExpectedColumn = iParen*iIndentStep + iIndent + iIndentStep


def starts_with_paren(lTokens):
    iToken = utils.find_next_non_whitespace_token(1, lTokens)
    if isinstance(lTokens[iToken], parser.open_parenthesis):
        return True
    return False


def remove_non_arrays(lToi):
    lReturn = []
    for oToi in lToi:
        lTokens = oToi.get_tokens()
        if starts_with_paren(lTokens):
            lReturn.append(oToi)
    return lReturn


def remove_single_line_assignments(lToi):
    lReturn = []
    for oToi in lToi:
        if oToi.iFirstLine != oToi.iLastLine:
            lReturn.append(oToi)
    return lReturn


def set_last_line_number(oToi):
    iLine, lTokens = utils.get_toi_parameters(oToi)
    for oToken in lTokens:

        iLine = utils.increment_line_number(iLine, oToken)

    oToi.iLastLine = iLine


def populate_toi_parameters(aToi, oFile):
    for oToi in aToi:
        oToi.iFirstLine = oToi.iLine
        lTemp = oFile.get_tokens_from_line(oToi.iLine)
        iIndent = len(lTemp.get_tokens()[0].get_value())
        oToi.iFirstLineIndent = iIndent

        oToi.iFirstLineIndentIndex = oFile.get_indent_of_line_at_index(oToi.get_start_index())

        oToi.iAssignColumn = oFile.get_column_of_token_index(oToi.get_start_index())
        set_last_line_number(oToi)


def check_indents(self, oToi, oLines):
    for oLine in oLines.lLines[1:]:
        sExpectedIndent = convert_column_index_to_whitespace(self, oLine.get_expected_indent(), oLines.get_first_line_indent(), oToi.iFirstLineIndentIndex)
        oLine.sExpectedIndent = sExpectedIndent
        if sExpectedIndent != oLine.actual_leading_whitespace:
            oViolation = create_violation(oToi, oLine)
            self.add_violation(oViolation)


def convert_column_index_to_whitespace(self, iColumn, iFirstLineIndent, iFirstLineIndentIndex):
    if self.indentStyle == 'smart_tabs':
        sIndent = '\t' * iFirstLineIndentIndex
        sAlignment = ' ' * (iColumn - iFirstLineIndent)
    else:
        sIndent = ' ' * self.indentSize * iFirstLineIndentIndex
        sAlignment = ' ' * (iColumn - len(sIndent))
    sLeadingWhitespace = sIndent + sAlignment
    return sLeadingWhitespace


def create_violation(oToi, oLine):
    iToken = oLine.token_index
    iLine = oLine.number
    lTokens = oToi.extract_tokens(iToken, iToken)
    sSolution = alignment_utils.build_solution(oLine.sExpectedIndent)

    oViolation = violation.New(iLine, lTokens, sSolution)
    oViolation.set_action(create_action_dict(oLine))

    return oViolation


def create_action_dict(oLine):
    dAction = {}
    dAction['line'] = oLine.number
    dAction['column'] = oLine.iExpectedIndent
    dAction['whitespace'] = oLine.sExpectedIndent

    if isinstance(oLine.tokens[0], parser.whitespace):
        dAction['action'] = 'adjust'
    else:
        dAction['action'] = 'insert'
    return dAction


class lines():

    def __init__(self, oToi):
        self.oToi = oToi
        self.lParens = []
        self.lLines = []
        self._convert_toi(oToi)

    def _convert_toi(self, oToi):
        iLine, lTokens = utils.get_toi_parameters(oToi)
        lLine = []
        bFirstLine = True
        iOffset = 0
        for iToken, oToken in enumerate(lTokens):
            lLine.append(oToken)
            if rules_utils.token_is_carriage_return(oToken):
                oLine = line(lLine, iLine, iToken - len(lLine) + 1, iOffset)
                oLine.isFirst = bFirstLine
                bFirstLine = False
                self.lLines.append(oLine)
                iLine += 1
                iOffset = 0
                lLine = []

        oLine = line(lLine, iLine, iToken - len(lLine) + 1)
        oLine.isLast = True
        self.lLines.append(oLine)

    def set_first_line_indent(self, iIndent):
        self.lLines[0].actual_indent = iIndent

    def get_first_line_indent(self):
        return self.lLines[0].actual_indent

    def set_first_line_expected_indent(self, iIndent):
        self.lLines[0].set_expected_indent(iIndent)

    def no_parens(self):
        if len(self.lParens) == 0:
            return True
        return False

    def update_parens(self, oLine, iIndentStep, iAdjust=0):
        self.lParens = oLine.update_parens(self.lParens, iIndentStep, iAdjust)


class line():

    def __init__(self, lLine, iLine, iToken, iIndent=0):
        self.number = iLine
        self.tokens = lLine
        self.parens = []
        self.populate_paren_list(iIndent)
        self.set_actual_indent()
        self.set_actual_leading_whitespace()
        self.token_index = iToken
        self.isFirst = False
        self.isLast = False
        self.iExpectedIndent = 0

    def populate_paren_list(self, iIndent):
        iColumn = iIndent
        for oToken in self.tokens:
            iColumn += len(oToken.get_value())
            if rules_utils.token_is_open_paren(oToken):
                oParen = paren('open', iColumn)
                self.parens.append(oParen)
            if rules_utils.token_is_close_paren(oToken):
                oParen = paren('close', iColumn)
                self.parens.append(oParen)

    def set_actual_indent(self):
        oToken = self.tokens[0]
        self.actual_indent = 0
        if rules_utils.token_is_whitespace(oToken):
            self.actual_indent = len(oToken.get_value())

    def set_actual_leading_whitespace(self):
        oToken = self.tokens[0]
        self.actual_leading_whitespace = ''
        if rules_utils.token_is_whitespace(oToken):
            self.actual_leading_whitespace = oToken.get_value()

    def get_delta_parens(self):
        iReturn = 0
        for oParen in self.parens:
            if oParen.is_open():
                iReturn += 1
            else:
                iReturn -= 1
        return iReturn

    def set_expected_indent(self, iIndent):
        self.iExpectedIndent = iIndent

    def get_expected_indent(self):
        return self.iExpectedIndent

    def starts_with_close_paren(self):
        return rules_utils.token_list_begins_with_close_paren(self.tokens)

    def update_parens(self, lParens, iIndentStep, iAdjust=0):
        update_parens_expected_column(self, iIndentStep, iAdjust)
        lParens.extend(self.parens)
        lReturn = remove_matching_parens(lParens)
        return lReturn


def update_parens_expected_column(self, iIndentStep, iAdjust=0):
    for oParen in self.parens:
        oParen.iExpectedColumn = oParen.iColumn + iAdjust - self.actual_indent + self.iExpectedIndent + iIndentStep - 1


def remove_matching_parens(lParens):
    lReturn = []
    for oParen in lParens:
        if oParen.is_open():
            lReturn.append(oParen)
        if oParen.is_close():
            lReturn.pop()
    return lReturn


class paren():

    def __init__(self, sType, iColumn, iId=None):
        self.sType = sType
        self.iColumn = iColumn
        self.iId = iId
        self.iExpectedColumn = iColumn

    def is_open(self):
        if self.sType == 'open':
            return True
        return False

    def is_close(self):
        if self.sType == 'close':
            return True
        return False


def align_left(self):
    if self.align_left == 'yes':
        return True
    return False


def align_paren(self):
    if self.align_paren == 'yes':
        return True
    return False




from vsg.rules import token_indent

from vsg import token

lTokens = []
lTokens.append(token.file_declaration.file_keyword)
lTokens.append(token.file_open_information.open_keyword)
lTokens.append(token.file_open_information.is_keyword)


class rule_001(token_indent):
    '''
    This rule checks the indent of **file** declarations.

    **Violation**

    .. code-block:: vhdl

       architecture rtl of fifo is

       file defaultImage : load_file_type open read_mode is load_file_name;

       file defaultImage : load_file_type open read_mode
       is load_file_name;

       begin

    **Fix**

    .. code-block:: vhdl

       architecture rtl of fifo is

         file defaultImage : load_file_type open read_mode is load_file_name;

         file defaultImage : load_file_type open read_mode
           is load_file_name;

       begin
    '''

    def __init__(self):
        token_indent.__init__(self, 'file', '001', lTokens)




from vsg import parser


class New():

    def __init__(self):
        self.code_tags = []
        self.next_line_code_tags = []
        self.bIgnoreNextCarriageReturn = False

    def clear(self):
        self.code_tags.clear()
        self.next_line_code_tags.clear()

    def clear_next_line_code_tags(self):
        self.next_line_code_tags.clear()

    def remove(self, sCodeTag):
        self.code_tags.remove(sCodeTag)

    def add(self, sCodeTag):
        if sCodeTag not in self.code_tags:
            self.code_tags.append(sCodeTag)

    def get_tags(self):
        lReturn = []
        lReturn.extend(self.code_tags)
        lReturn.extend(self.next_line_code_tags)
        return lReturn

    def update(self, oToken):

        if isinstance(oToken, parser.carriage_return):
            if self.bIgnoreNextCarriageReturn:
                self.bIgnoreNextCarriageReturn = False
            else:
                self.next_line_code_tags.clear()
            return None

        if on_code_tag_detected(oToken):
            remove_code_tags(self, oToken)
            return False
        elif off_code_tag_detected(oToken):
            add_code_tags(self, oToken)
            return True
        elif next_line_code_tag_detected(oToken):
            add_next_line_code_tags(self, oToken)
            self.bIgnoreNextCarriageReturn = True


def on_code_tag_detected(oToken):
    if not isinstance(oToken, parser.comment):
        return False
    if token_starts_with(oToken, '-- vsg_on'):
        return True
    return False


def off_code_tag_detected(oToken):
    if not isinstance(oToken, parser.comment):
        return False
    if token_starts_with(oToken, '-- vsg_off'):
        return True
    return False


def next_line_code_tag_detected(oToken):
    if not isinstance(oToken, parser.comment):
        return False
    if token_starts_with(oToken, '-- vsg_disable_next_line'):
        return True
    return False


def token_starts_with(oToken, sString):
    if oToken.get_value().startswith(sString):
        return True
    return False


def remove_code_tags(self, oToken):
    sValue = remove_code_tag_comment(oToken)
    lValues = sValue.split()
    if bare_code_tag(lValues):
        self.clear()
    else:
       for sCodeTag in lValues[2:]:
           self.remove(sCodeTag)


def add_code_tags(self, oToken):
    sValue = remove_code_tag_comment(oToken)
    lValues = sValue.split()
    if bare_code_tag(lValues):
        self.clear()
        self.add('all')
    else:
       for sCodeTag in lValues[2:]:
           self.add(sCodeTag)


def remove_code_tag_comment(oToken):
    lString = oToken.get_value().split(':')
    return lString[0]


def add_next_line_code_tags(self, oToken):
    sValue = remove_code_tag_comment(oToken)
    lValues = sValue.split()
    for sCodeTag in lValues[2:]:
       if sCodeTag not in self.next_line_code_tags:
           self.next_line_code_tags.append(sCodeTag)


def bare_code_tag(lValues):
    if len(lValues) == 2:
        return True
    return False


def token_has_vsg_on_code_tag(oToken):
    return on_code_tag_detected(oToken)


def token_has_vsg_off_code_tag(oToken):
    return off_code_tag_detected(oToken)


def token_has_next_line_code_tag(oToken):
    return next_line_code_tag_detected(oToken)




from vsg.rules import token_case

from vsg import token

lTokens = []
lTokens.append(token.subprogram_kind.function_keyword)


class rule_014(token_case):
    '''
    This rule checks the **function** keyword in the **end function** has proper case.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end FUNCTION;

       end Function foo;

    **Fix**

    .. code-block:: vhdl

       end function;

       end function foo;
    '''

    def __init__(self):
        token_case.__init__(self, 'function', '014', lTokens)
        self.groups.append('case::keyword')





from vsg.rules import move_token_next_to_another_token as Rule

from vsg.token import full_type_declaration as token


class rule_017(Rule):
    '''
    This rule checks the identifier is on the same line as the type keyword.

    **Violation**

    .. code-block:: vhdl

       type
       t_record is

    **Fix**

    .. code-block:: vhdl

       type t_record
       is
    '''

    def __init__(self):
        Rule.__init__(self, 'type', '017', token.type_keyword, token.identifier)
        self.solution = 'Ensure identifier is on the same line as the *type* keyword.'




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003

from .rule_100 import rule_100
from .rule_101 import rule_101

from .rule_200 import rule_200
from .rule_201 import rule_201
from .rule_202 import rule_202
from .rule_203 import rule_203

from .rule_300 import rule_300
from .rule_301 import rule_301

from .rule_400 import rule_400
from .rule_401 import rule_401
from .rule_402 import rule_402

from .rule_500 import rule_500
from .rule_501 import rule_501
from .rule_502 import rule_502
from .rule_503 import rule_503
from .rule_504 import rule_504
from .rule_505 import rule_505
from .rule_506 import rule_506
from .rule_507 import rule_507

from .rule_600 import rule_600
from .rule_601 import rule_601




from vsg.rules import align_tokens_in_region_between_tokens

from vsg import token

lAlign = []
lAlign.append(token.interface_unknown_declaration.colon)


class rule_017(align_tokens_in_region_between_tokens):
    '''
    This rule checks the alignment of the colon for each generic and port in the component declaration.

    Following extra configurations are supported:

    * :code:`separate_generic_port_alignment`.

    |configuring_keyword_alignment_rules_link|

    **Violation**

    .. code-block:: vhdl

       component my_component
           generic (
               g_width : positive;
               g_output_delay : positive
           );
           port (
               clk_i : in std_logic;
               data_i : in std_logic;
               data_o : in std_logic
           );
       end component;

    **Fix**

    .. code-block:: vhdl

       component my_component
           generic (
               g_width        : positive;
               g_output_delay : positive
           );
           port (
               clk_i  : in std_logic;
               data_i : in std_logic;
               data_o : in std_logic
           );
       end component;
    '''

    def __init__(self):
        align_tokens_in_region_between_tokens.__init__(self, 'component', '017', lAlign, token.component_declaration.component_keyword, token.component_declaration.end_keyword)
        self.solution = 'Align :.'




from vsg import parser
from vsg import token

from vsg.rules.whitespace_between_token_pairs_bounded_by_tokens import Rule

lTokens = []
lTokens.append([token.interface_constant_declaration.assignment, parser.todo])
lTokens.append([token.interface_signal_declaration.assignment, parser.todo])
lTokens.append([token.interface_variable_declaration.assignment, parser.todo])
lTokens.append([token.interface_unknown_declaration.assignment, parser.todo])

oStart = token.generic_clause.open_parenthesis
oEnd = token.generic_clause.close_parenthesis


class rule_006(Rule):
    '''
    This rule checks for a single space after the default assignment.

    **Violation**

    .. code-block:: vhdl

       g_width : integer :=32;
       g_depth : integer :=     512;

    **Fix**

    .. code-block:: vhdl

       g_width : integer := 32;
       g_depth : integer := 512;
    '''
    def __init__(self):
        Rule.__init__(self, 'generic', '006', lTokens, oStart, oEnd)




from .rule_001 import rule_001
from .rule_002 import rule_002
from .rule_003 import rule_003

from .rule_100 import rule_100

from .rule_300 import rule_300
from .rule_301 import rule_301
from .rule_302 import rule_302

from .rule_400 import rule_400
from .rule_401 import rule_401

from .rule_500 import rule_500
from .rule_501 import rule_501




from vsg.token import file_open_information as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import file_logical_name


def detect(iToken, lObjects):
    '''
    file_open_information ::=
        [ open *file_open_kind*_expression ] is file_logical_name
    '''

    if utils.is_next_token_one_of(['open', 'is'], iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = iToken

    if utils.is_next_token('open', iCurrent, lObjects):
        iCurrent = utils.assign_next_token_required('open', token.open_keyword, iCurrent, lObjects)
        iCurrent = utils.assign_next_token(token.file_open_kind_expression, iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('is', token.is_keyword, iCurrent, lObjects)

    iCurrent = file_logical_name.classify(iCurrent, lObjects)

    return iCurrent




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.case_generate_statement.generate_label)
lTokens.append(token.for_generate_statement.generate_label)
lTokens.append(token.if_generate_statement.generate_label)


class rule_004(previous_line):
    '''
    This rule checks for blank lines or comments before the **generate** label.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <= '1';
       ram_array : for i in 0 to 7 generate

    **Fix**

    .. code-block:: vhdl

       wr_en <= '1';

       ram_array : for i in 0 to 7 generate
    '''

    def __init__(self):
        previous_line.__init__(self, 'generate', '004', lTokens)




from vsg import parser


class attribute_keyword(parser.keyword):
    '''
    unique_id = attribute_declaration : attribute_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class identifier(parser.identifier):
    '''
    unique_id = attribute_declaration : identifier
    '''

    def __init__(self, sString):
        parser.identifier.__init__(self, sString)


class colon(parser.colon):
    '''
    unique_id = attribute_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class semicolon(parser.semicolon):
    '''
    unique_id = attribute_declaration : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment as Rule

from vsg.token import record_type_definition as token

lTokens = []
lTokens.append(token.record_keyword)


class rule_002(Rule):
    '''
    This rule checks for code after the **record** keyword.

    **Violation**

    .. code-block:: vhdl

       type t_record is record a : std_logic;
         b : std_logic;
       end record;

    **Fix**

    .. code-block:: vhdl

       type t_record is record
         a : std_logic;
         b : std_logic;
       end record;
    '''

    def __init__(self):
        Rule.__init__(self, 'record_type_definition', '002', lTokens)
        self.solution = 'Move code after the begin to the next line.'




from vsg import deprecated_rule


class rule_015(deprecated_rule.Rule):
    '''
    This rule has been deprecated.
    The **component** keyword is required per the LRM.
    '''

    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'component', '015')
        self.message.append('Rule ' + self.unique_id + ' is required per the LRM.')




from vsg.token import interface_package_generic_map_aspect as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_map_aspect


def detect(iToken, lObjects):
    '''
    interface_package_generic_map_aspect ::=
        generic_map_aspect
      | generic map ( <> )
      | generic map ( default )
    '''
    if utils.are_next_consecutive_tokens(['generic', 'map', '(', '<>'], iToken, lObjects):
        return classify(iToken, lObjects)
    elif utils.are_next_consecutive_tokens(['generic', 'map', '(', 'default'], iToken, lObjects):
        return classify(iToken, lObjects)
    else:
       return generic_map_aspect.classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('generic', token.generic_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('map', token.map_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('default', token.default_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('<>', token.undefined_range, iToken, lObjects)
    iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iToken, lObjects)

    return iCurrent




from vsg import parser
from vsg import token
from vsg import violation

from vsg.rules import utils as rules_utils
from vsg.rule_group import structure
from vsg.vhdlFile import utils

lMoveTokens = []
lMoveTokens.append(token.conditional_expressions.else_keyword)
lMoveTokens.append(token.conditional_waveforms.else_keyword)

lTokenPairs = []
lTokenPairs.append([token.conditional_force_assignment.assignment, token.conditional_force_assignment.semicolon])
lTokenPairs.append([token.conditional_variable_assignment.assignment, token.conditional_variable_assignment.semicolon])
lTokenPairs.append([token.concurrent_conditional_signal_assignment.assignment, token.concurrent_conditional_signal_assignment.semicolon])
lTokenPairs.append([token.conditional_waveform_assignment.assignment, token.conditional_waveform_assignment.semicolon])


class rule_001(structure.Rule):
    '''
    This rule checks the **else** keyword is not at the beginning of a line.
    The else should be at the end of the preceeding line.

    **Violation**

    .. code-block:: vhdl

       wr_en <= '1' when a = '1' -- This is comment
                else '0' when b = '0'
                else c when d = '1'
                else f;

    **Fix**

    .. code-block:: vhdl

       wr_en <= '1' when a = '1' else -- This is a comment
                '0' when b = '0' else
                c when d = '1' else
                f;
    '''

    def __init__(self):
        structure.Rule.__init__(self, 'when', '001')
        self.lMoveTokens = lMoveTokens
        self.lTokenPairs = lTokenPairs

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1])
            lToi = utils.combine_two_token_class_lists(lToi, aToi)
        return lToi

    def _analyze(self, lToi):
        for oToi in lToi:
            lTokens = oToi.get_tokens()
            iLine = oToi.get_line_number()
            oAnchorToken = None
            iMoveToLine = None
            iLeft = None

            for iToken, oToken in enumerate(lTokens):
                iLine = utils.increment_line_number(iLine, oToken)
                for oMoveToken in self.lMoveTokens:
                    if isinstance(oToken, oMoveToken):
                        if (lTokens[iToken - 2] is not oAnchorToken and lTokens[iToken - 1] is not oAnchorToken) or \
                           lTokens[iToken - 2] is oAnchorToken and isinstance(lTokens[iToken - 1], parser.carriage_return):
                            self.solution = 'Move "' + oToken.get_value() + '" to the right of "' + oAnchorToken.get_value() + '" on line ' + str(iMoveToLine)
                            if isinstance(lTokens[iToken + 1], parser.whitespace):
                                iRight = iToken + 1
                            else:
                                iRight = iToken
                            oViolation = violation.New(iLine, oToi.extract_tokens(iLeft, iRight), self.solution)
                            self.add_violation(oViolation)
                            break
                if not utils.token_is_whitespace_or_comment(oToken):
                    iLeft = iToken + 1
                    iMoveToLine = iLine
                    oAnchorToken = oToken

    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        if isinstance(lTokens[-1], parser.whitespace):
            lTokens.pop()
        oToken = lTokens.pop()
        rules_utils.insert_token(lTokens, 0, oToken)
        rules_utils.insert_whitespace(lTokens, 0)
        oViolation.set_tokens(lTokens)




from vsg.token import parameter_specification as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import discrete_range


def classify_until(lUntils, iToken, lObjects):
    '''
    parameter_specification ::=
        identifier in discrete_range
    '''

    iCurrent = utils.assign_next_token(token.identifier, iToken, lObjects)
    iCurrent = utils.assign_next_token_required('in', token.in_keyword, iCurrent, lObjects)

    iCurrent = discrete_range.classify_until(lUntils, iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_case_n_token_after_tokens_between_tokens

from vsg import token

lTokens = []
lTokens.append(token.interface_constant_declaration.colon)
lTokens.append(token.interface_variable_declaration.colon)
lTokens.append(token.interface_signal_declaration.colon)
lTokens.append(token.interface_unknown_declaration.colon)

oStart = token.port_clause.open_parenthesis
oEnd = token.port_clause.close_parenthesis


class rule_018(token_case_n_token_after_tokens_between_tokens):
    '''
    This rule checks the port type has proper case if it is a VHDL keyword.

    |configuring_uppercase_and_lowercase_rules_link|


    **Violation**

    .. code-block:: vhdl

       port (
         WR_EN    : in    STD_LOGIC;
         RD_EN    : in    std_logic;
         OVERFLOW : out   t_OVERFLOW;
         DATA     : inout STD_LOGIC_VECTOR(31 downto 0)
       );

    **Fix**

    .. code-block:: vhdl

       port (
         WR_EN    : in    std_logic;
         RD_EN    : in    std_logic;
         OVERFLOW : out   t_OVERFLOW;
         DATA     : inout std_logic_vector(31 downto 0)
       );
    '''

    def __init__(self):
        token_case_n_token_after_tokens_between_tokens.__init__(self, 'port', '018', 2, lTokens, oStart, oEnd, True)
        self.disabled = True
        self.groups.append('case::keyword')




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.instantiated_unit.component_keyword, token.instantiated_unit.component_name])


class rule_032(Rule):
    '''
    This rule checks for a single space after the **component** keyword if it is used.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       INSTANCE_NAME : component ENTITY_NAME
       INSTANCE_NAME : component   ENTITY_NAME
       INSTANCE_NAME : component  ENTITY_NAME

    **Fix**

    .. code-block:: vhdl

       INSTANCE_NAME : component ENTITY_NAME
       INSTANCE_NAME : component ENTITY_NAME
       INSTANCE_NAME : component ENTITY_NAME
    '''
    def __init__(self):
        Rule.__init__(self, 'instantiation', '032', lTokens)




from vsg import token

from vsg.rules.whitespace_after_token import Rule
from vsg.rules import utils

lTokens = []
lTokens.append(token.interface_constant_declaration.colon)
lTokens.append(token.interface_signal_declaration.colon)
lTokens.append(token.interface_variable_declaration.colon)
lTokens.append(token.interface_file_declaration.colon)
lTokens.append(token.interface_unknown_declaration.colon)

oStart = token.generic_clause.open_parenthesis
oEnd = token.generic_clause.close_parenthesis


class rule_005(Rule):
    '''
    This rule checks for a single space after the colon in a generic declaration.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       g_width :integer := 32;

    **Fix**

    .. code-block:: vhdl

       g_width : integer := 32;
    '''
    def __init__(self):
        Rule.__init__(self, 'generic', '005', lTokens)
        self.solution = 'Reduce number of spaces after the colon to 1.'

    def _get_tokens_of_interest(self, oFile):
        lToi = oFile.get_token_and_n_tokens_after_it_when_between_tokens(self.lTokens, 2, oStart, oEnd)
        lToi = utils.remove_toi_if_token_is_at_the_end_of_the_line(lToi)
        return lToi




from vsg.vhdlFile.classify import range_constraint


def detect(iToken, lObjects):
    '''
    integer_type_definition ::=
        range_constraint
    '''

    return range_constraint.detect(iToken, lObjects)




from vsg.rules.blank_line_above_line_starting_with_token_when_between_tokens import Rule

from vsg.token import generic_clause as token
from vsg.token import entity_declaration as between

lTokens = []
lTokens.append(token.generic_keyword)


class rule_200(Rule):
    '''
    This rule checks for blank lines above the **generic** keyword in entity specifications.

    **Violation**

    .. code-block:: vhdl

       entity fifo is



         generic (

    **Fix**

    .. code-block:: vhdl

       entity fifo is
         generic (
    '''
    def __init__(self):
        Rule.__init__(self, 'entity', '200', lTokens)
        self.style = 'no_blank_line'
        self.lBetweenTokenPairs = [between.entity_keyword, between.semicolon]




from vsg import deprecated_rule


class rule_020(deprecated_rule.Rule):
    '''
    This rule has been renamed to `port_map_005 <port_map_rules.html#port-map-005>`_.
    '''
    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'instantiation', '020')
        self.message.append('Rule ' + self.unique_id + ' has been renamed to rule port_map_005.')




from vsg import deprecated_rule


class rule_502(deprecated_rule.Rule):
    '''
    This rule has been deprecated.  The case of *entity_designator* should be enforced by other rules.
    '''
    def __init__(self):
        deprecated_rule.Rule.__init__(self, 'entity_specification', '502')
        self.message.append('Rule ' + self.unique_id + ' has been deprecated.')




from vsg.token import force_mode as token

from vsg.vhdlFile import utils


def detect(iToken, lObjects):
    '''
    force_mode ::=
        in | out
    '''

    iCurrent = utils.assign_next_token_if('in', token.in_keyword, iToken, lObjects)
    iCurrent = utils.assign_next_token_if('out', token.out_keyword, iCurrent, lObjects)

    return iCurrent




from vsg.vhdlFile.extract import utils

from vsg.vhdlFile.extract.get_line_preceeding_line import get_line_preceeding_line


def get_line_above_line_starting_with_token(lTokens, lAllTokens, oTokenMap, bIncludeComments=False):

    lReturn = []

    lTokenIndexes = utils.get_indexes_of_token_list(lTokens, oTokenMap)
    lIndexes = []
    for iIndex in lTokenIndexes:
        if utils.is_token_at_start_of_line(iIndex, oTokenMap):
            lIndexes.append(iIndex)

    lLine = utils.get_line_numbers_of_indexes_in_list(lIndexes, oTokenMap)

    for iLine in lLine:
        lReturn.append(get_line_preceeding_line(iLine, lAllTokens, 1, oTokenMap, bIncludeComments))

    return lReturn




from vsg.vhdlFile.classify import expression


def classify(iToken, lObjects):
    '''
    file_logical_name ::= *string*_expression
    '''

    iCurrent = expression.classify_until([';'], iToken, lObjects)

    return iCurrent




from vsg import parser


class colon(parser.colon):
    '''
    unique_id = element_declaration : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)


class semicolon(parser.semicolon):
    '''
    unique_id = element_declaration : semicolon
    '''

    def __init__(self, sString=';'):
        parser.semicolon.__init__(self)




from vsg.vhdlFile import utils

from vsg.token import component_specification as token

from vsg.vhdlFile.classify import instantiation_list


def detect(iToken, lObjects):
    iCurrent = iToken

    while utils.find_in_next_n_tokens(',', 2, iCurrent, lObjects):
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        iCurrent = utils.find_next_token(iCurrent + 1, lObjects)

    if utils.find_in_next_n_tokens(':', 2, iCurrent, lObjects):
        return True
    return False


def classify(iToken, lObjects):
    '''
    component_specification ::=
        instantiation_list : component_name
    '''

    iCurrent = iToken

    iCurrent = instantiation_list.classify(iToken, lObjects)

    iCurrent = utils.assign_next_token_required(':', token.colon, iCurrent, lObjects)
    iCurrent = utils.assign_next_token(token.component_name, iCurrent, lObjects)

    return iCurrent




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.subprogram_body.end_keyword, token.subprogram_kind.procedure_keyword])
lTokens.append([token.subprogram_body.end_keyword, token.subprogram_body.designator])
lTokens.append([token.subprogram_kind.procedure_keyword, token.subprogram_body.designator])


class rule_101(Rule):
    '''
    This rule checks for a single space between the **end** and **procedure** keywords and procedure designator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       end   procedure   average_samples;
       end   procedure;
       end   average_samples;

    **Fix**

    .. code-block:: vhdl

       end procedure average_samples;
       end procedure;
       end average_samples;
    '''
    def __init__(self):
        Rule.__init__(self, 'procedure', '101', lTokens)




from vsg.token import sensitivity_clause as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import sensitivity_list


def detect(iToken, lObjects):
    '''
    sensitivity_clause ::=
        on sensitivity_list
    '''

    if utils.is_next_token('on', iToken, lObjects):
        return True
    return False


def classify_until(lUntils, iToken, lObjects):

    iCurrent = utils.assign_next_token_required('on', token.on_keyword, iToken, lObjects)

    iCurrent = sensitivity_list.classify_until(lUntils, iCurrent, lObjects)

    return iCurrent




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.process_statement.process_keyword)
lTokens.append(token.process_statement.process_label)


class rule_015(previous_line):
    '''
    This rule checks for blank lines or comments above the **process** declaration.

    |configuring_blank_lines_link|

    The default style is :code:`no_code`.

    **Violation**

    .. code-block:: vhdl

       -- This process performs FIFO operations.
       proc_a : process (rd_en, wr_en, data_in, data_out,

       wr_en <= wr_en;
       proc_a : process (rd_en, wr_en, data_in, data_out,

    **Fix**

    .. code-block:: vhdl

       -- This process performs FIFO operations.
       proc_a : process (rd_en, wr_en, data_in, data_out,

       wr_en <= wr_en;

       proc_a : process (rd_en, wr_en, data_in, data_out,
    '''

    def __init__(self):
        previous_line.__init__(self, 'process', '015', lTokens)
        self.style = 'no_code'




from vsg.rules import move_token as Rule

from vsg import token

oToken = token.generic_clause.close_parenthesis


class rule_010(Rule):
    '''
    This rule checks the location of the closing ")" character for the generic clause.

    The default location is on a line by itself.

    |configuring_move_token_rules_link|

    **Violation**

    .. code-block:: vhdl

       g_depth : integer := 512);

    **Fix**

    .. code-block:: vhdl

         g_depth : integer := 512
       );
    '''

    def __init__(self):
        Rule.__init__(self, 'generic', '010', oToken)
        self.preserve_comment = True




from vsg import parser

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import range


def detect(iToken, lObjects):
    '''
    discrete_range ::=
        *discrete*_subtype_indication | range
    '''
    return range.detect(iToken, lObjects)


def classify(iToken, lObjects):
    '''
    discrete_range ::=
        *discrete*_subtype_indication | range
    '''

    return utils.assign_token(lObjects, iToken, parser.todo)


def classify_until(lUntils, iToken, lObjects):
    iCurrent = iToken
    iStop = len(lObjects) - 1
    iOpenParenthesis = 0
    iCloseParenthesis = 0
    while iCurrent < iStop:
        iCurrent = utils.find_next_token(iCurrent, lObjects)
        if utils.token_is_open_parenthesis(iCurrent, lObjects):
           iOpenParenthesis += 1
        if utils.token_is_close_parenthesis(iCurrent, lObjects):
           iCloseParenthesis += 1
        if iOpenParenthesis < iCloseParenthesis:
            break
        elif iOpenParenthesis == iCloseParenthesis:
            if lObjects[iCurrent].get_value().lower() in lUntils:
                break
            else:
                utils.assign_token(lObjects, iCurrent, parser.todo)
        else:
            utils.assign_token(lObjects, iCurrent, parser.todo)
    return iCurrent




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.loop_statement.loop_label)


class rule_503(token_case_with_prefix_suffix):
    '''
    This rule checks the proper case of the label on a loop statement.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

         LABEL : for index in 4 to 23 loop
         Label : for index in 0 to 100 loop

    **Fix**

    .. code-block:: vhdl

         label : for index in 4 to 23 loop
         label : for index in 0 to 100 loop
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'loop_statement', '503', lTokens)
        self.groups.append('case::label')




from vsg.rules import token_suffix

from vsg import token

lTokens = []
lTokens.append(token.signal_declaration.identifier)


class rule_600(token_suffix):
    '''
    This rule checks for valid suffixes on signal identifiers.
    Default signal suffix is *\_s*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       signal wr_en : std_logic;
       signal rd_en : std_logic;

    **Fix**

    .. code-block:: vhdl

       signal wr_en_s : std_logic;
       signal rd_en_s : std_logic;
    '''

    def __init__(self):
        token_suffix.__init__(self, 'signal', '600', lTokens)
        self.suffixes = ['_s']
        self.solution = 'Signal identifiers'




from vsg.rules import remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace

from vsg.token import report_statement as token


class rule_001(remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace):
    '''
    This rule removes labels on report_statement_statements.

    **Violation**

    .. code-block:: vhdl

        REPORT_LABEL : report "FIFO width is limited to 16 bits.";

    **Fix**

    .. code-block:: vhdl

        report "FIFO width is limited to 16 bits.";
    '''

    def __init__(self):
        remove_tokens_bounded_by_tokens_and_remove_trailing_whitespace.__init__(self, 'report_statement', '001', token.label, token.label_colon)
        self.solution = 'Remove Label'




from vsg.vhdlFile.classify import configuration_declaration
from vsg.vhdlFile.classify import context_declaration
from vsg.vhdlFile.classify import entity_declaration
from vsg.vhdlFile.classify import package_declaration
from vsg.vhdlFile.classify import package_instantiation_declaration


def detect(iToken, lObjects):
    '''
    primary_unit ::=
        entity_declaration
      | configuration_declaration
      | package_declaration
      | package_instantiation_declaration
      | context_declaration
      | PSL_Verification_Unit
    '''

    iReturned = context_declaration.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    iReturned = entity_declaration.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    iReturned = package_declaration.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    iReturned = package_instantiation_declaration.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    iReturned = configuration_declaration.detect(iToken, lObjects)
    if iReturned != iToken:
        return iReturned

    return iToken




from vsg.token import case_statement_alternative as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import choices
from vsg.vhdlFile.classify import sequence_of_statements


def detect(iToken, lObjects):
    '''
    case_statement_alternative ::=
        when choices =>
            sequence_of_statements
    '''
    if utils.is_next_token('when', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    iCurrent = utils.assign_next_token_required('when', token.when_keyword, iToken, lObjects)

    iCurrent = choices.classify_until(['=>'], iCurrent, lObjects)

    iCurrent = utils.assign_next_token_required('=>', token.assignment, iCurrent, lObjects)

    iCurrent = sequence_of_statements.detect(iCurrent, lObjects)

    return iCurrent




from vsg.rules import token_case_with_prefix_suffix

from vsg import token

lTokens = []
lTokens.append(token.component_declaration.component_simple_name)


class rule_012(token_case_with_prefix_suffix):
    '''
    This rule checks the proper case of the component name in the **end component** line.

    |configuring_uppercase_and_lowercase_rules_link|

    **Violation**

    .. code-block:: vhdl

       end component FIFO;

    **Fix**

    .. code-block:: vhdl

       end component fifo;
    '''

    def __init__(self):
        token_case_with_prefix_suffix.__init__(self, 'component', '012', lTokens)
        self.groups.append('case::name')




from vsg.rules import line_length


class rule_001(line_length):
    '''
    This rule checks the length of the line.

    |configuring_length_rules_link|
    '''

    def __init__(self):
        line_length.__init__(self, 'length', '001')




from vsg import token

from vsg.rules import token_prefix

lTokens = []
lTokens.append(token.component_instantiation_statement.instantiation_label)


class rule_601(token_prefix):
    '''
    This rule checks for valid prefixes on instantiation labels.
    The default prefix is *inst_*.

    |configuring_prefix_and_suffix_rules_link|

    **Violation**

    .. code-block:: vhdl

       fifo_32x2k : FIFO

    **Fix**

    .. code-block:: vhdl

       inst_fifo_32x2k : FIFO
    '''

    def __init__(self):
        token_prefix.__init__(self, 'instantiation', '601', lTokens)
        self.prefixes = ['inst_']
        self.solution = 'instantiation label'




from vsg.rules import insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens

from vsg import token


class rule_016(insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens):
    '''
    This rule checks for multiple generics defined on a single line.

    **Violation**

    .. code-block:: vhdl

      generic (
        g_width : std_logic := '0';g_depth : std_logic := '1'
      );

    **Fix**

    .. code-block:: vhdl

      generic (
        g_width : std_logic := '0';
        g_depth : std_logic := '1'
      );
    '''

    def __init__(self):
        insert_carriage_return_after_token_if_it_is_not_followed_by_a_comment_when_between_tokens.__init__(self, 'generic', '016', [token.interface_list.semicolon], token.generic_clause.open_parenthesis, token.generic_clause.close_parenthesis)
        self.solution = 'Move multiple generics to their own lines.'




from vsg import token
from vsg import parser
from vsg import violation

from vsg.rules import alignment_utils
from vsg.rules import utils as rules_utils
from vsg.rule_group import alignment
from vsg.vhdlFile import utils


class multiline_alignment_between_tokens(alignment.Rule):
    '''
    Checks the case for words.

    Parameters
    ----------

    name : string
       The group the rule belongs to.

    identifier : string
       unique identifier.  Usually in the form of 00N.

    trigger : parser object type
       object type to apply the case check against
    '''

    def __init__(self, name, identifier, lTokenPairs, bExcludeLastToken=False):
        alignment.Rule.__init__(self, name=name, identifier=identifier)
        self.phase = 4
        self.lTokenPairs = lTokenPairs
        self.bExcludeLastToken = bExcludeLastToken
        self.align_left = 'no'
        self.configuration.append('align_left')
        self.align_paren = 'yes'
        self.configuration.append('align_paren')
        self.bIgnoreStartParen = False
        self.bConstraint = False
        self.iIndentAfterParen = 1

    def _get_tokens_of_interest(self, oFile):
        lToi = []
        for lTokenPair in self.lTokenPairs:
            aToi = oFile.get_tokens_bounded_by(lTokenPair[0], lTokenPair[1], bExcludeLastToken=self.bExcludeLastToken)
            lToi = utils.combine_two_token_class_lists(lToi, aToi)

        lReturn = []
        for oToi in lToi:
            if toi_is_an_array(oToi):
                continue
            iLine, lTokens = utils.get_toi_parameters(oToi)
            iFirstLine, iFirstLineIndent = alignment_utils.get_first_line_info(iLine, oFile)
            iAssignColumn = oFile.get_column_of_token_index(oToi.get_start_index())
            oToi.set_meta_data('iFirstLine', iFirstLine)
            oToi.set_meta_data('iFirstLineIndent', iFirstLineIndent)
            oToi.set_meta_data('iAssignColumn', iAssignColumn)
            oToi.set_meta_data('bStartsWithParen', alignment_utils.starts_with_paren(lTokens))
            lReturn.append(oToi)

        return lReturn

    def _analyze(self, lToi):

        for oToi in lToi:
            iLine, lTokens = utils.get_toi_parameters(oToi)

            iFirstLine = oToi.get_meta_data('iFirstLine')
            iFirstLineIndent = oToi.get_meta_data('iFirstLineIndent')
            iAssignColumn = oToi.get_meta_data('iAssignColumn')
            bStartsWithParen = oToi.get_meta_data('bStartsWithParen')

            iColumn = iAssignColumn

            dActualIndent = {}

            dActualIndent[iLine] = iFirstLineIndent
            lParens = []
            dIndex = {}

            bSkipCommentLine = False

            for iToken, oToken in enumerate(lTokens):

                iLine = utils.increment_line_number(iLine, oToken)

                if isinstance(oToken, parser.blank_line):
                    continue

                if bSkipCommentLine:
                   if not isinstance(oToken, parser.carriage_return):
                       continue

                if isinstance(oToken, parser.carriage_return):
                    iColumn = 0
                    bSkipCommentLine = rules_utils.does_line_start_with_comment(lTokens[iToken + 1:iToken + 3])
                    if bSkipCommentLine:
                        dActualIndent[iLine] = None
                    else:
                        dActualIndent[iLine] = alignment_utils.set_indent(iToken, lTokens)
                        dIndex[iLine] = iToken + 1
                    continue

                iColumn += len(oToken.get_value())

                if isinstance(oToken, parser.close_parenthesis):
                    dParen = {}
                    dParen['type'] = 'close'
                    dParen['line'] = iLine
                    dParen['column'] = iColumn
                    dParen['begin_line'] = utils.does_token_start_line(iToken, lTokens)
                    lParens.append(dParen)

                if isinstance(oToken, parser.open_parenthesis):
                    dParen = {}
                    dParen['type'] = 'open'
                    dParen['line'] = iLine
                    dParen['column'] = iColumn
                    lParens.append(dParen)

            iLastLine = iLine

            if iFirstLine == iLastLine:
                continue

            iFirstTokenLength = len(lTokens[0].get_value())

            if self.align_paren == 'no' and self.align_left == 'yes':
                dExpectedIndent = _analyze_align_left_yes_align_paren_no(iFirstLine, iLastLine, lParens, self.indentSize, dActualIndent, bStartsWithParen, self.bIgnoreStartParen)
            if self.align_paren == 'yes' and self.align_left == 'no':
                dExpectedIndent = _analyze_align_left_no_align_paren_yes(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn, iFirstTokenLength, self.bIgnoreStartParen, self.iIndentAfterParen)
            if self.align_paren == 'yes' and self.align_left == 'yes':
                dExpectedIndent = _analyze_align_paren_yes_align_left_yes(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn, self.bIgnoreStartParen, self.bConstraint)
            if self.align_paren == 'no' and self.align_left == 'no':
                dExpectedIndent = _analyze_align_paren_no_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, self.indentSize, bStartsWithParen, iAssignColumn, self.bIgnoreStartParen)


#            print(f'Actual = {dActualIndent}')
#            print(f'Expect = {dExpectedIndent}')
#            print(f'Index  = {dIndex}')

            if self.indentStyle == 'smart_tabs':
                alignment_utils.convert_expected_indent_to_smart_tab(dExpectedIndent, self.indentSize, iFirstLineIndent)

            for iLine in range(iFirstLine, iLastLine + 1):
                if dActualIndent[iLine] is None:
                    continue
                if indents_match(dActualIndent[iLine], dExpectedIndent[iLine]):
                    continue

                oViolation = build_violation(iLine, oToi, iToken, dExpectedIndent, dIndex, dActualIndent)
                self.add_violation(oViolation)


    def _fix_violation(self, oViolation):
        lTokens = oViolation.get_tokens()
        dAction = oViolation.get_action()
#        print(dAction)
        if dAction['action'] == 'adjust':
            lTokens[0].set_value(dAction['column'])
        else:
            rules_utils.insert_new_whitespace(lTokens, 0, dAction['column'])

        oViolation.set_tokens(lTokens)


def build_violation(iLine, oToi, iToken, dExpectedIndent, dIndex, dActualIndent):
    sSolution = alignment_utils.build_solution(dExpectedIndent[iLine])
    dAction = build_action_dictionary(iLine, dActualIndent, dExpectedIndent)
    iToken = dIndex[iLine]
    oViolation = violation.New(iLine, oToi.extract_tokens(iToken, iToken), sSolution)
    oViolation.set_action(dAction)
    return oViolation


def build_action_dictionary(iLine, dActualIndent, dExpectedIndent):
    dAction = {}
    dAction['line'] = iLine
    dAction['column'] = dExpectedIndent[iLine]

    if dActualIndent[iLine] == '':
        dAction['action'] = 'insert'
    else:
        dAction['action'] = 'adjust'
    return dAction


def line_starts_with_comment(iActual):
    if iActual is None:
        return True
    return False


def indents_match(iActual, iExpected):
    if iActual == iExpected:
        return True
    return False


def calculate_start_column(oFile, oToi):
    iReturn = oFile.get_column_of_token_index(oToi.get_start_index())
    iReturn += len(oToi.get_tokens()[0].get_value())
    iReturn += 1
#    print(f'Start Column = {iReturn}')
    return iReturn


def is_token_before_carriage_return(tToken, lTokens):
    for oToken in lTokens:
        if isinstance(oToken, tToken):
            return True
        if isinstance(oToken, parser.carriage_return):
            return False
    return False


def _set_column_adjustment(iToken, lTokens):
    iReturn = 0
    if isinstance(lTokens[iToken + 1], parser.whitespace):
        if isinstance(lTokens[iToken + 2], parser.close_parenthesis):
            iReturn = -1
    else:
        if isinstance(lTokens[iToken + 1], parser.close_parenthesis):
            iReturn = -1
    return iReturn


def _analyze_align_left_yes_align_paren_no(iFirstLine, iLastLine, lParens, iIndentStep, dActualIndent, bStartsWithParen, bIgnoreStartParen):
#    print('--> _analyze_align_left_yes_align_paren_no <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]
    if bStartsWithParen or bIgnoreStartParen:
        iFirstIndent = dActualIndent[iFirstLine]
    else:
        iFirstIndent = dActualIndent[iFirstLine] + iIndentStep

    iIndent = iFirstIndent

#    print(iIndent)
    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')

        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = iIndent
            continue

        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
            else:
                iParens -= 1
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] -= iIndentStep

        iIndent = iFirstIndent + iParens * iIndentStep

#        print(f'indent = {iIndent} | iPerens = {iParens}')
        dExpectedIndent[iLine + 1] = iIndent
#        print(dExpectedIndent)

    dReturn = {}
    dReturn[iFirstLine] = dExpectedIndent[iFirstLine]
    for iLine in range(iFirstLine + 1, iLastLine + 1):

        dReturn[iLine] = dExpectedIndent[iLine] * ' '

    return dReturn


def _analyze_align_left_no_align_paren_yes(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn, iFirstTokenLength, bIgnoreStartParen, iIndentAfterParen):
#    print('--> _analyze_align_left_no_align_paren_yes <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

    if bIgnoreStartParen:
        iIndent = iAssignColumn + iFirstTokenLength
    else:
        iIndent = iAssignColumn + iFirstTokenLength + 1

    iColumn = iIndent
    lColumn = [iIndent]

#    print('*'*80)
#    print(lParens)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <' + '-'*80)
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column']
#                    print(f"iColumn = {iColumn}")
                else:
                    iColumn = dTemp['column'] + iTemp - len(dActualIndent[iLine]) + iIndentStep - 1
                    if iParens == 1:
#                        print('Got Here')
                        iColumn -= iIndentAfterParen 
#                    print(f"iColumn = {dTemp['column']} + ({iTemp} - {len({dActualIndent[iLine]})}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                if bIgnoreStartParen:
                    dExpectedIndent[iLine + 1] = iColumn + 1
                else:
                    dExpectedIndent[iLine + 1] = iColumn
            else:
                iParens -= 1
                lColumn.pop()
                if bIgnoreStartParen:
                    dExpectedIndent[iLine + 1] = (lColumn[-1] + 1)
                else:
                    dExpectedIndent[iLine + 1] = lColumn[-1]
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = dExpectedIndent[iLine] - iIndentStep
#                    dExpectedIndent[iLine] = dExpectedIndent[iLine] - 2

#            print(f'lColumn = {lColumn}')

#        print(f'iParens = {iParens}')


        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = lColumn[-1]

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')
#    print(dExpectedIndent)
#    print(convert_numbers_to_spaces(dExpectedIndent))
    return convert_numbers_to_spaces(dExpectedIndent)


def convert_numbers_to_spaces(dExpectedIndent):
    dReturn = {}
    lKeys = list(dExpectedIndent.keys())
    dReturn[lKeys[0]] = dExpectedIndent[lKeys[0]]
    for sKey in lKeys[1:]:
        dReturn[sKey] = ' ' * dExpectedIndent[sKey]
    return dReturn

def _analyze_align_paren_yes_align_left_yes(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn, bIgnoreStartParen, bConstraint):
#    print('--> _analyze_align_paren_yes_align_left_yes <-' + '-'*70)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]

    if bStartsWithParen:
       iIndent = dActualIndent[iFirstLine]
       iColumn = iIndent
       lColumn = [dActualIndent[iFirstLine]]
    else:
#       if bIgnoreStartParen:
#           iIndent = dActualIndent[iFirstLine]
#       else:
       iIndent = iAssignColumn + 2 + 1
       iColumn = iIndent
       lColumn = [iIndent]
#    print(iIndent)
#    print('*'*80)
#    print(lParens)

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')
        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

#        print(f'lTemp = {lTemp}')
        iTemp = lColumn[-1]
        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
                if iLine == iFirstLine:
                    iColumn = dTemp['column'] + iIndentStep - 1
                else:
                    iColumn = dTemp['column'] + (iTemp - len(dActualIndent[iLine])) + iIndentStep - 1
#                print(f"iColumn = {dTemp['column']} + ({iTemp} - {dActualIndent[iLine]}) + {iIndentStep} - 1 = {iColumn}")
                lColumn.append(iColumn)
                dExpectedIndent[iLine + 1] = iColumn
            else:
                iParens -= 1
                lColumn.pop()
                dExpectedIndent[iLine + 1] = lColumn[-1]
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] = dExpectedIndent[iLine] - iIndentStep

#        print(f'iParens = {iParens}')

        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = lColumn[-1]

        if iLine == iFirstLine:
            if bConstraint:
                dExpectedIndent[iLine + 1] = iParens * iIndentStep + dActualIndent[iFirstLine] + iIndentStep
            else:
                dExpectedIndent[iLine + 1] = lColumn[-1] - 1
            lColumn[-1] = iParens * iIndentStep + dActualIndent[iFirstLine] + iIndentStep
            if iParens == 0 and not bStartsWithParen:
                dExpectedIndent[iLine + 1] = iIndentStep + dActualIndent[iFirstLine]
                lColumn[-1] = iIndentStep + dActualIndent[iFirstLine]
            elif iParens == 0 and bStartsWithParen:
                dExpectedIndent[iLine + 1] = dActualIndent[iFirstLine]
                lColumn[-1] = dActualIndent[iFirstLine]
            if bIgnoreStartParen:
                dExpectedIndent[iLine + 1] = dActualIndent[iFirstLine] + iIndentStep
                lColumn[-1] = dActualIndent[iFirstLine] + iIndentStep

        else:
            if bConstraint:
                if iParens == 1:
                    dExpectedIndent[iLine + 1] = iParens * iIndentStep + dActualIndent[iFirstLine]
                    lColumn[-1] = iParens * iIndentStep + dActualIndent[iFirstLine]
            else:
                if iParens == 0:
                    dExpectedIndent[iLine + 1] = lColumn[-1]
                else:
                    dExpectedIndent[iLine + 1] -= 1

#        print(f'{iLine} | {lColumn} | {dExpectedIndent}')

    return convert_numbers_to_spaces(dExpectedIndent)


def _analyze_align_paren_no_align_left_no(iFirstLine, iLastLine, lParens, dActualIndent, iIndentStep, bStartsWithParen, iAssignColumn, bIgnoreStartParen):
#    print('--> _analyze_align_paren_no <-' + '-'*70)
#    print(iAssignColumn)
#    print(lParens)
    dExpectedIndent = {}
    dExpectedIndent[iFirstLine] = dActualIndent[iFirstLine]
#    print(iIndent)
    if bStartsWithParen:
        iFirstIndent = dActualIndent[iFirstLine]
    elif bIgnoreStartParen:
        iFirstIndent = lParens[0]['column'] + 1
    else:
        iFirstIndent = iAssignColumn + 2 + 1

    iIndent = iFirstIndent

    iParens = 0

    for iLine in range(iFirstLine, iLastLine + 1):
#        print('-->  ' + str(iLine) + '  <--------------------------')

        lTemp = []
        for dParen in lParens:
            if dParen['line'] == iLine:
                lTemp.append(dParen)

        if len(lTemp) == 0:
            dExpectedIndent[iLine + 1] = iIndent
            continue

        for dTemp in lTemp:
            if dTemp['type'] == 'open':
                iParens += 1
            else:
                iParens -= 1
                if dTemp['begin_line']:
                    dExpectedIndent[iLine] -= iIndentStep

        if iLine == iFirstLine:
            if bStartsWithParen:
                iIndent = iFirstIndent + iParens * iIndentStep
            elif bIgnoreStartParen:
                iIndent = iFirstIndent
            else:
                if iParens == 0:
                    iIndent = iFirstIndent
                else:
                    iIndent = iFirstIndent + iParens * iIndentStep
        else:
            if iParens == 0:
                iIndent = iFirstIndent + iParens * iIndentStep
            elif bIgnoreStartParen:
                iIndent = iFirstIndent + iParens * iIndentStep - iIndentStep
            else:
                iIndent = iFirstIndent + iParens * iIndentStep

#        print(f'indent = {iIndent} | iPerens = {iParens}')
        dExpectedIndent[iLine + 1] = iIndent

    dReturn = {}
    dReturn[iFirstLine] = dExpectedIndent[iFirstLine]
    for iLine in range(iFirstLine + 1, iLastLine + 1):

        dReturn[iLine] = dExpectedIndent[iLine] * ' '

    return dReturn


def toi_is_an_array(oToi):
    lTokens = oToi.get_tokens()
    for oToken in lTokens:
        if isinstance(oToken, token.aggregate.open_parenthesis):
            return True
    return False




from vsg import parser


class entity_class(parser.keyword):
    '''
    unique_id = entity_specification : entity_class
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class colon(parser.colon):
    '''
    unique_id = entity_specification : colon
    '''

    def __init__(self, sString=':'):
        parser.colon.__init__(self)




from vsg.rules.whitespace_after_token import Rule

from vsg import token

lTokens = []
lTokens.append(token.concurrent_simple_signal_assignment.assignment)
lTokens.append(token.concurrent_conditional_signal_assignment.assignment)
lTokens.append(token.concurrent_selected_signal_assignment.assignment)


class rule_002(Rule):
    '''
    This rule checks for a single space after the **<=** operator.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       wr_en <=    '0';
       rd_en <=   '1';

    **Fix**

    .. code-block:: vhdl

       wr_en <= '0';
       rd_en <= '1';
    '''

    def __init__(self):
        Rule.__init__(self, 'concurrent', '002', lTokens)




from vsg.token import concurrent_procedure_call_statement as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import procedure_call


def detect(iToken, lObjects):
    '''
    concurrent_procedure_call_statement ::=
        [ label : ] [ postponed ] procedure_call ;
    '''
    iCurrent = iToken
    if procedure_call.detect(iToken, lObjects):
        iCurrent = utils.tokenize_label(iToken, lObjects, token.label_name, token.label_colon)
        iCurrent = utils.tokenize_postponed(iCurrent, lObjects, token.postponed_keyword)
        iCurrent = procedure_call.classify(iCurrent, lObjects)
        iCurrent = utils.assign_next_token_required(';', token.semicolon, iCurrent, lObjects)

    return iCurrent




from vsg.rules import previous_line

from vsg import token

lTokens = []
lTokens.append(token.block_statement.block_label)


class rule_200(previous_line):
    '''
    This rule checks for blank lines or comments above the block label.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       a <= b;
       block_label : block is

    **Fix**

    .. code-block:: vhdl

       a <= b;

       block_label : block is
    '''

    def __init__(self):
        previous_line.__init__(self, 'block', '200', lTokens)




from vsg.rules import previous_line

from vsg.token import loop_statement as token

lTokens = []
lTokens.append(token.end_keyword)


class rule_202(previous_line):
    '''
    This rule checks for blank lines or comments above the **end** keyword.

    |configuring_previous_line_rules_link|

    **Violation**

    .. code-block:: vhdl

       loop
         a <= b;
       end loop;

    **Fix**

    .. code-block:: vhdl

       loop
         a <= b;

       end loop;
    '''
    def __init__(self):
        previous_line.__init__(self, 'loop_statement', '202', lTokens)




from vsg.rules import split_line_at_token

from vsg import token

lTokens = []
lTokens.append(token.block_statement.begin_keyword)


class rule_004(split_line_at_token):
    '''
    This rule checks the **begin** keyword is on its own line.

    **Violation**

    .. code-block:: vhdl

       block is begin

    **Fix**

    .. code-block:: vhdl

       block is
       begin
    '''

    def __init__(self):
        split_line_at_token.__init__(self, 'block', '004', lTokens)
        self.solution = 'Move *begin* keyword and code after it to the next line'




from vsg.token import subprogram_header as token

from vsg.vhdlFile import utils

from vsg.vhdlFile.classify import generic_list
from vsg.vhdlFile.classify import generic_map_aspect


def detect(iToken, lObjects):
    '''
    subprogram_header ::=
        [ generic ( generic_list )
        [ generic_map_aspect ] ]
    '''

    if utils.is_next_token('generic', iToken, lObjects):
        return classify(iToken, lObjects)
    return iToken


def classify(iToken, lObjects):

    if utils.find_in_next_n_tokens('(', 2, iToken, lObjects):
        iCurrent = utils.assign_next_token_required('generic', token.generic_keyword, iToken, lObjects)
        iCurrent = utils.assign_next_token_required('(', token.open_parenthesis, iCurrent, lObjects)

        iCurrent = generic_list.classify(iCurrent, lObjects)

        iCurrent = utils.assign_next_token_required(')', token.close_parenthesis, iCurrent, lObjects)

    iCurrent = generic_map_aspect.detect(iToken, lObjects)

    return iCurrent




from vsg import parser


class after_keyword(parser.keyword):
    '''
    unique_id = waveform_element : after_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)


class null_keyword(parser.keyword):
    '''
    unique_id = waveform_element : null_keyword
    '''

    def __init__(self, sString):
        parser.keyword.__init__(self, sString)




from vsg import token

from vsg.rules.whitespace_between_token_pairs import Rule

lTokens = []
lTokens.append([token.library_clause.keyword, token.logical_name_list.logical_name])


class rule_002(Rule):
    '''
    This rule checks for excessive spaces after the **library** keyword.

    |configuring_whitespace_rules_link|

    **Violation**

    .. code-block:: vhdl

       library    ieee;

    **Fix**

    .. code-block:: vhdl

       library ieee;
    '''
    def __init__(self):
        Rule.__init__(self, 'library', '002', lTokens)




from vsg import token

from vsg.rules import token_suffix

lTokens = []
lTokens.append(token.case_generate_statement.generate_label)
lTokens.append(token.case_generate_statement.end_generate_label)
lTokens.append(token.for_generate_statement.generate_label)
lTokens.append(token.for_generate_statement.end_generate_label)
lTokens.append(token.if_generate_statement.generate_label)
lTokens.append(token.if_generate_statement.end_generate_label)


class rule_600(token_suffix):
    '''
    This rule checks for valid suffixes on generate statement labels.
    The default suffix is *\_gen*.

    |configuring_prefix_and_suffix_rules_link|

    [Violation]

       label : case condition generate

    [Fix]

       label_gen : case condition generate
    '''

    def __init__(self):
        token_suffix.__init__(self, 'generate', '600', lTokens)
        self.suffixes = ['_gen']
